"""TokenizationMode enum and the MazeTokenizer class"""

from enum import Enum
from functools import cached_property
from typing import Callable, Iterable, Mapping, Sequence

import numpy as np
from muutils.json_serialize import (
    SerializableDataclass,
    serializable_dataclass,
    serializable_field,
)
from muutils.kappa import Kappa

from maze_dataset.constants import SPECIAL_TOKENS, CoordTup
from maze_dataset.tokenization.util import (
    _coord_to_strings_indexed,
    _coord_to_strings_UT,
    coords_to_strings,
    strings_to_coords,
)
from maze_dataset.utils import WhenMissing, corner_first_ndindex


class TokenError(ValueError):
    """error for tokenization"""

    pass


class TokenizationMode(Enum):
    """mode of tokenization

    # Abbreviations:
    - `AOTP`: Ajacency list, Origin, Target, Path
    - `UT`: Unique Token (for each coordiate)
    - `CTT`: Coordinate Tuple Tokens (each coordinate is tokenized as a tuple of integers)

    # Modes:
    - `AOTP_UT_rasterized`: the "classic" mode: assigning tokens to each coordinate is done via rasterization
        example: for a 3x3 maze, token order is `(0,0), (0,1), (0,2), (1,0), (1,1), (1,2), (2,0), (2,1), (2,2)`
    - `AOTP_UT_uniform`: new mode, where a 3x3 tokenization scheme and 5x5 tokenizations scheme are compatible
        uses `corner_first_ndindex` function to order the tokens
    - `AOTP_CTT_indexed`: each coordinate is a tuple of integers
    """

    AOTP_UT_rasterized = "AOTP_UT_rasterized"
    AOTP_UT_uniform = "AOTP_UT_uniform"
    AOTP_CTT_indexed = "AOTP_CTT_indexed"


_NDINDEX_FUNC_MAP: dict[
    TokenizationMode, Callable[[int], Iterable[tuple[int, ...]]]
] = {
    TokenizationMode.AOTP_UT_rasterized: lambda n: list(np.ndindex(n, n)),
    TokenizationMode.AOTP_UT_uniform: lambda n: corner_first_ndindex(n, 2),
}


def is_UT(tokenization_mode: TokenizationMode) -> bool:
    return tokenization_mode in (
        TokenizationMode.AOTP_UT_rasterized,
        TokenizationMode.AOTP_UT_uniform,
    )


_MAZETOKENIZER_PROPERTIES_TO_SERIALIZE: list[str] = [
    "name",
    "max_grid_size",
    "token_arr",
    "tokenizer_map",
    "vocab_size",
    "padding_token_index",
]


@serializable_dataclass(
    properties_to_serialize=_MAZETOKENIZER_PROPERTIES_TO_SERIALIZE, kw_only=True
)
class MazeTokenizer(SerializableDataclass):
    """Tokenizer for mazes

    # Parameters:
     - `tokenization_mode: TokenizationMode`
        mode of tokenization. required.
    - `max_grid_size: int | None`
        maximum grid size. required for actually turning text tokens to numerical tokens, but not for moving between coordinates/mazes and text

    # Properties
    - `name: str`
        auto-generated name of the tokenizer from mode and size

    ## Conditional Properties

    - `node_strings_map: Mapping[CoordTup, str]`
        map from node to string. This returns a `muutils.kappa.Kappa` object which you can use like a dictionary. returns `None` if not a `UT` mode

    these all return `None` if `max_grid_size` is `None`.
    Prepend `_` to the name to get a guaranteed type, and cause an exception if `max_grid_size` is `None`

    - `token_arr: list[str]`
        list of tokens, in order of their indices in the vocabulary
    - `tokenizer_map: Mapping[str, int]`
        map from token to index
    - `vocab_size: int`
        size of the vocabulary
    - `padding_token_index: int`
        index of the padding token

    # Methods
    - `coords_to_strings(coords: list[CoordTup]) -> list[str]`
        convert a list of coordinates to a list of tokens. Optionally except, skip, or ignore non-coordinates
    - `strings_to_coords(strings: list[str]) -> list[CoordTup]`
        convert a list of tokens to a list of coordinates. Optionally except, skip, or ignore non-coordinates

    """

    # parameters
    # ============================================================

    tokenization_mode: TokenizationMode = serializable_field(
        default=TokenizationMode.AOTP_UT_uniform,
        serialization_fn=lambda x: x.value,
        loading_fn=lambda x: TokenizationMode[x["tokenization_mode"]],
    )

    max_grid_size: int | None = serializable_field(default=None)

    # properties
    # ============================================================

    @property
    def name(self) -> str:
        max_grid_size_str: str = (
            f"-g{self.max_grid_size}" if self.max_grid_size is not None else ""
        )
        return f"maze_tokenizer-{self.tokenization_mode.value}{max_grid_size_str}"

    @cached_property
    def _node_strings_map(self) -> Mapping[CoordTup, list[str]]:
        """map a coordinate to a token"""
        if self.tokenization_mode in (
            TokenizationMode.AOTP_UT_rasterized,
            TokenizationMode.AOTP_UT_uniform,
        ):
            return Kappa(_coord_to_strings_UT)
        elif self.tokenization_mode == TokenizationMode.AOTP_CTT_indexed:
            return Kappa(_coord_to_strings_indexed)
        else:
            raise ValueError(
                f"Invalid tokenization mode {self.tokenization_mode}",
                f"expected one of {TokenizationMode.__members__}",
            )

    @cached_property
    def node_strings_map(self) -> Mapping[CoordTup, list[str]] | None:
        """map a coordinate to a token"""
        if self.tokenization_mode in (
            TokenizationMode.AOTP_UT_rasterized,
            TokenizationMode.AOTP_UT_uniform,
        ):
            return None
        else:
            return self._node_strings_map

    # conditional properties (on max_grid_size existing)
    # ------------------------------------------------------------

    @cached_property
    def _token_arr(self) -> list[str]:
        """map from index to token"""
        if self.max_grid_size is None:
            raise ValueError(
                f"max_grid_size must be specified to use token_arr property: {self.max_grid_size = }"
            )

        output: list[str] = list(SPECIAL_TOKENS.values())

        if self.tokenization_mode in (
            TokenizationMode.AOTP_UT_rasterized,
            TokenizationMode.AOTP_UT_uniform,
        ):
            output.extend(
                [
                    self._node_strings_map[coord][0]
                    for coord in _NDINDEX_FUNC_MAP[self.tokenization_mode](
                        self.max_grid_size
                    )
                ]
            )
        elif self.tokenization_mode == TokenizationMode.AOTP_CTT_indexed:
            # TODO: this is hacky, but we don't want to modify the original SPECIAL_TOKENS since that will break old models
            output.extend(
                [
                    "(",
                    ",",
                    ")",  # new special chars
                    *map(str, range(self.max_grid_size)),  # numbers
                ]
            )
        else:
            raise ValueError(
                f"Invalid tokenization mode {self.tokenization_mode}",
                f"expected one of {TokenizationMode.__members__}",
            )

        return output

    @cached_property
    def token_arr(self) -> list[str] | None:
        if self.max_grid_size is None:
            return None
        return self._token_arr

    @cached_property
    def _tokenizer_map(self) -> dict[str, int]:
        """map from token to index"""
        return {token: i for i, token in enumerate(self._token_arr)}

    @cached_property
    def tokenizer_map(self) -> dict[str, int] | None:
        if self.max_grid_size is None:
            return None
        return self._tokenizer_map

    @property
    def _vocab_size(self) -> int:
        return len(self._token_arr)

    @property
    def vocab_size(self) -> int | None:
        if self.max_grid_size is None:
            return None
        return self._vocab_size

    @property
    def _n_tokens(self) -> int:
        # TODO: deprecate
        return self._vocab_size

    @property
    def n_tokens(self) -> int | None:
        if self.max_grid_size is None:
            return None
        return self._n_tokens

    @cached_property
    def _padding_token_index(self) -> int:
        return self.tokenizer_map[SPECIAL_TOKENS.PADDING]

    @cached_property
    def padding_token_index(self) -> int | None:
        if self.max_grid_size is None:
            return None
        return self._padding_token_index

    # conversion functions
    # ============================================================

    def coords_to_strings(
        self,
        coords: list[CoordTup],
        when_noncoord: WhenMissing = "skip",
    ) -> list[str]:
        if self.tokenization_mode in (
            TokenizationMode.AOTP_UT_rasterized,
            TokenizationMode.AOTP_UT_uniform,
        ):
            return coords_to_strings(
                coords=coords,
                coord_to_strings_func=_coord_to_strings_UT,
                when_noncoord=when_noncoord,
            )
        elif self.tokenization_mode == TokenizationMode.AOTP_CTT_indexed:
            return coords_to_strings(
                coords=coords,
                coord_to_strings_func=_coord_to_strings_indexed,
                when_noncoord=when_noncoord,
            )
        else:
            raise ValueError(
                f"Invalid tokenization mode {self.tokenization_mode}",
                f"expected one of {TokenizationMode.__members__}",
            )

    @staticmethod
    def strings_to_coords(
        text: str,
        when_noncoord: WhenMissing = "skip",
    ) -> list[str | CoordTup]:
        return strings_to_coords(text=text, when_noncoord=when_noncoord)

    def encode(self, text: str | list[str]) -> list[int]:
        """encode a string or list of strings into a list of tokens"""
        try:
            if isinstance(text, str):
                text = text.split()
            return [self.tokenizer_map[token] for token in text]
        except KeyError as e:
            raise TokenError(
                f"Token {e} not found",
                f"in vocabulary of {self}:",
                f"{self.token_arr}",
            ) from e

    def decode(
        self, tokens: Sequence[int], joined_tokens: bool = False
    ) -> list[str] | str:
        """decode a list of tokens into a string or list of strings"""
        try:
            output: list[str] = [self.token_arr[token] for token in tokens]
        except IndexError as e:
            raise TokenError(
                f"Token index '{e}' not found in vocabulary of length {self.vocab_size}"
            ) from e
        if joined_tokens:
            return " ".join(output)
        else:
            return output

    # UT-only coordinate stuff
    # ============================================================

    @cached_property
    def coordinate_tokens_coords(self) -> dict[CoordTup, int]:
        print(f"{self.tokenization_mode = }")
        if not self.is_UT():
            raise ValueError(
                f"coordinate_tokens_coords is only valid for UT tokenization modes, got {self.tokenization_mode = }"
            )
        if self.max_grid_size is None:
            raise ValueError(
                f"max_grid_size must be specified to use coordinate_tokens: {self.max_grid_size = }"
            )

        raw_converted: list[CoordTup | str] = self.strings_to_coords(
            self.token_arr, when_noncoord="include"
        )

        # filter out non-coordinates
        return {
            coord: i
            for i, coord in enumerate(raw_converted)
            if not isinstance(coord, str)
        }

    @cached_property
    def coordinate_tokens_ids(self) -> dict[str, int]:
        # checks performed in call
        output: dict[str, int] = dict()

        for coord, index in self.coordinate_tokens_coords.items():
            _for_key: list[str] = self.coords_to_strings([coord])
            assert len(_for_key) == 1
            output[_for_key[0]] = index

        return output

    # other
    # ============================================================

    def summary(self) -> dict:
        """returns a summary of the tokenization mode"""
        return {
            "tokenization_mode": self.tokenization_mode.value,
            "max_grid_size": self.max_grid_size,
            "vocab_size": self.vocab_size,
        }

    def is_AOTP(self) -> bool:
        """returns true if a tokenization mode is Adjacency list, Origin, Target, Path"""
        return self.tokenization_mode in (
            TokenizationMode.AOTP_UT_rasterized,
            TokenizationMode.AOTP_UT_uniform,
            TokenizationMode.AOTP_CTT_indexed,
        )

    def is_UT(self) -> bool:
        return is_UT(self.tokenization_mode)

    def clear_cache(self):
        """clears all cached properties"""
        # delete the properties only if they exist
        for name, prop in self.__class__.__dict__.items():
            if isinstance(prop, cached_property):
                # if the property exists, delete it
                try:
                    delattr(self, name)
                except AttributeError as e:
                    pass



import typing
import warnings
from dataclasses import dataclass
from itertools import chain

import numpy as np
from jaxtyping import Bool, Float, Int, Int8, Shaped
from muutils.json_serialize.serializable_dataclass import (
    SerializableDataclass,
    serializable_dataclass,
    serializable_field,
)
from muutils.misc import list_split

from maze_dataset.constants import (
    NEIGHBORS_MASK,
    SPECIAL_TOKENS,
    Coord,
    CoordArray,
    CoordList,
    CoordTup,
)
from maze_dataset.tokenization import (
    MazeTokenizer,
    TokenizationMode,
    get_adj_list_tokens,
    get_path_tokens,
)
from maze_dataset.tokenization.token_utils import get_origin_tokens, get_target_tokens

ConnectionList = Bool[np.ndarray, "lattice_dim x y"]
RGB = tuple[int, int, int]

PixelGrid = Int[np.ndarray, "x y rgb"]
BinaryPixelGrid = Bool[np.ndarray, "x y"]


def _fill_edges_with_walls(connection_list: ConnectionList) -> ConnectionList:
    """fill the last elements of the connections lists as false for each dim"""
    for dim in range(connection_list.shape[0]):
        # last row for down
        if dim == 0:
            connection_list[dim, -1, :] = False
        # last column for right
        elif dim == 1:
            connection_list[dim, :, -1] = False
        else:
            raise NotImplementedError(f"only 2d lattices supported. got {dim=}")
    return connection_list


def color_in_pixel_grid(pixel_grid: PixelGrid, color: RGB) -> bool:
    for row in pixel_grid:
        for pixel in row:
            if np.all(pixel == color):
                return True
    return False


@dataclass(frozen=True)
class PixelColors:
    WALL: RGB = (0, 0, 0)
    OPEN: RGB = (255, 255, 255)
    START: RGB = (0, 255, 0)
    END: RGB = (255, 0, 0)
    PATH: RGB = (0, 0, 255)


@dataclass(frozen=True)
class AsciiChars:
    WALL: str = "#"
    OPEN: str = " "
    START: str = "S"
    END: str = "E"
    PATH: str = "X"


ASCII_PIXEL_PAIRINGS: dict[str, RGB] = {
    AsciiChars.WALL: PixelColors.WALL,
    AsciiChars.OPEN: PixelColors.OPEN,
    AsciiChars.START: PixelColors.START,
    AsciiChars.END: PixelColors.END,
    AsciiChars.PATH: PixelColors.PATH,
}


@serializable_dataclass(
    frozen=True,
    kw_only=True,
    properties_to_serialize=["lattice_dim", "generation_meta"],
)
class LatticeMaze(SerializableDataclass):
    """lattice maze (nodes on a lattice, connections only to neighboring nodes)

    Connection List represents which nodes (N) are connected in each direction.

    First and second elements represent rightward and downward connections,
    respectively.

    Example:
      Connection list:
        [
          [ # down
            [F T],
            [F F]
          ],
          [ # right
            [T F],
            [T F]
          ]
        ]

      Nodes with connections
        N T N F
        F   T
        N T N F
        F   F

      Graph:
        N - N
            |
        N - N

    Note: the bottom row connections going down, and the
    right-hand connections going right, will always be False.
    """

    connection_list: ConnectionList
    generation_meta: dict | None = serializable_field(default=None, compare=False)

    lattice_dim = property(lambda self: self.connection_list.shape[0])
    grid_shape = property(lambda self: self.connection_list.shape[1:])
    n_connections = property(lambda self: self.connection_list.sum())

    @property
    def grid_n(self) -> int:
        assert self.grid_shape[0] == self.grid_shape[1], "only square mazes supported"
        return self.grid_shape[0]

    # ============================================================
    # basic methods
    # ============================================================
    @staticmethod
    def heuristic(a: CoordTup, b: CoordTup) -> float:
        """return manhattan distance between two points"""
        return np.abs(a[0] - b[0]) + np.abs(a[1] - b[1])

    def __hash__(self) -> int:
        return hash(self.connection_list.tobytes())

    def nodes_connected(self, a: Coord, b: Coord, /) -> bool:
        """returns whether two nodes are connected"""
        delta: Coord = b - a
        if np.abs(delta).sum() != 1:
            # return false if not even adjacent
            return False
        else:
            # test for wall
            dim: int = int(np.argmax(np.abs(delta)))
            clist_node: Coord = a if (delta.sum() > 0) else b
            return self.connection_list[dim, clist_node[0], clist_node[1]]

    def is_valid_path(self, path: CoordArray, empty_is_valid: bool = False) -> bool:
        """check if a path is valid"""
        # check path is not empty
        if len(path) == 0:
            if not empty_is_valid:
                return False
            else:
                return True

        # check all coords in bounds of maze
        if not np.all((0 <= path) & (path < self.grid_shape)):
            return False

        # check all nodes connected
        for i in range(len(path) - 1):
            if not self.nodes_connected(path[i], path[i + 1]):
                return False
        return True

    def get_coord_neighbors(self, c: Coord) -> CoordArray:
        neighbors: list[Coord] = [
            neighbor
            for neighbor in (c + NEIGHBORS_MASK)
            if (
                (0 <= neighbor[0] < self.grid_shape[0])  # in x bounds
                and (0 <= neighbor[1] < self.grid_shape[1])  # in y bounds
                and self.nodes_connected(c, neighbor)  # connected
            )
        ]

        output: CoordArray = np.array(neighbors)
        if len(neighbors) > 0:
            assert output.shape == (
                len(neighbors),
                2,
            ), f"invalid shape: {output.shape}, expected ({len(neighbors)}, 2))\n{c = }\n{neighbors = }\n{self.as_ascii()}"
        return output

    def gen_connected_component_from(self, c: Coord) -> CoordArray:
        """return the connected component from a given coordinate"""
        # Stack for DFS
        stack: list[Coord] = [c]

        # Set to store visited nodes
        visited: set[CoordTup] = set()

        while stack:
            current_node: Coord = stack.pop()
            # this is fine since we know current_node is a coord and thus of length 2
            visited.add(tuple(current_node))  # type: ignore[arg-type]

            # Get the neighbors of the current node
            neighbors = self.get_coord_neighbors(current_node)

            # Iterate over neighbors
            for neighbor in neighbors:
                if tuple(neighbor) not in visited:
                    stack.append(neighbor)

        return np.array(list(visited))

    def find_shortest_path(
        self,
        c_start: CoordTup,
        c_end: CoordTup,
    ) -> CoordArray:
        """find the shortest path between two coordinates, using A*"""
        c_start = tuple(c_start)
        c_end = tuple(c_end)

        g_score: dict[CoordTup, float] = (
            dict()
        )  # cost of cheapest path to node from start currently known
        f_score: dict[CoordTup, float] = {
            c_start: 0.0
        }  # estimated total cost of path thru a node: f_score[c] := g_score[c] + heuristic(c, c_end)

        # init
        g_score[c_start] = 0.0
        g_score[c_start] = self.heuristic(c_start, c_end)

        closed_vtx: set[CoordTup] = set()  # nodes already evaluated
        open_vtx: set[CoordTup] = set([c_start])  # nodes to be evaluated
        source: dict[CoordTup, CoordTup] = (
            dict()
        )  # node immediately preceding each node in the path (currently known shortest path)

        while open_vtx:
            # get lowest f_score node
            c_current: CoordTup = min(open_vtx, key=lambda c: f_score[c])
            # f_current: float = f_score[c_current]

            # check if goal is reached
            if c_end == c_current:
                path: list[CoordTup] = [c_current]
                p_current: CoordTup = c_current
                while p_current in source:
                    p_current = source[p_current]
                    path.append(p_current)
                # ----------------------------------------------------------------------
                # this is the only return statement
                return np.array(path[::-1])
                # ----------------------------------------------------------------------

            # close current node
            closed_vtx.add(c_current)
            open_vtx.remove(c_current)

            # update g_score of neighbors
            _np_neighbor: Coord
            for _np_neighbor in self.get_coord_neighbors(c_current):
                neighbor: CoordTup = tuple(_np_neighbor)

                if neighbor in closed_vtx:
                    # already checked
                    continue
                g_temp: float = g_score[c_current] + 1  # always 1 for maze neighbors

                if neighbor not in open_vtx:
                    # found new vtx, so add
                    open_vtx.add(neighbor)

                elif g_temp >= g_score[neighbor]:
                    # if already knew about this one, but current g_score is worse, skip
                    continue

                # store g_score and source
                source[neighbor] = c_current
                g_score[neighbor] = g_temp
                f_score[neighbor] = g_score[neighbor] + self.heuristic(neighbor, c_end)

        raise ValueError(
            "A solution could not be found!",
            f"{c_start = }, {c_end = }",
            self.as_ascii(),
        )

    def get_nodes(self) -> CoordArray:
        """return a list of all nodes in the maze"""
        rows: Int[np.ndarray, "x y"]
        cols: Int[np.ndarray, "x y"]
        rows, cols = np.meshgrid(
            range(self.grid_shape[0]),
            range(self.grid_shape[1]),
            indexing="ij",
        )
        nodes: CoordArray = np.vstack((rows.ravel(), cols.ravel())).T
        return nodes

    def get_connected_component(self) -> CoordArray:
        """get the largest (and assumed only nonsingular) connected component of the maze

        TODO: other connected components?
        """
        if (self.generation_meta is None) or (
            self.generation_meta.get("fully_connected", False)
        ):
            # for fully connected case, pick any two positions
            return self.get_nodes()
        else:
            # if metadata provided, use visited cells
            visited_cells: set[CoordTup] | None = self.generation_meta.get(
                "visited_cells", None
            )
            if visited_cells is None:
                # TODO: dynamically generate visited_cells?
                raise ValueError(
                    f"a maze which is not marked as fully connected must have a visited_cells field in its generation_meta: {self.generation_meta}\n{self}\n{self.as_ascii()}"
                )
            else:
                visited_cells_np: Int[np.ndarray, "N 2"] = np.array(list(visited_cells))
                return visited_cells_np

    def generate_random_path(
        self,
        except_when_invalid: bool = True,
        allowed_start: CoordList | None = None,
        allowed_end: CoordList | None = None,
        deadend_start: bool = False,
        deadend_end: bool = False,
    ) -> CoordArray:
        """return a path between randomly chosen start and end nodes within the connected component

        Note that setting special conditions on start and end positions might cause the same position to be selected as both start and end.

        # Parameters:
         - `except_when_invalid : bool`
            deprecated. setting this to `False` will cause an error.
           (defaults to `True`)
         - `allowed_start : CoordList | None`
            a list of allowed start positions. If `None`, any position in the connected component is allowed
           (defaults to `None`)
         - `allowed_end : CoordList | None`
            a list of allowed end positions. If `None`, any position in the connected component is allowed
           (defaults to `None`)
         - `deadend_start : bool`
            whether to ***force*** the start position to be a deadend (defaults to `False`)
           (defaults to `False`)
         - `deadend_end : bool`
            whether to ***force*** the end position to be a deadend (defaults to `False`)
           (defaults to `False`)

        # Returns:
         - `CoordArray`
            a path between the selected start and end positions

        # Raises:
         - `ValueError` : if the connected component has less than 2 nodes and `except_when_invalid` is `True`
        """

        # we can't create a "path" in a single-node maze
        assert (
            self.grid_shape[0] > 1 and self.grid_shape[1] > 1
        ), f"can't create path in single-node maze: {self.as_ascii()}"

        # get connected component
        connected_component: CoordArray = self.get_connected_component()

        # initialize start and end positions
        positions: Int[np.int8, "2 2"]

        # handle deprecated parameter
        if not except_when_invalid:
            raise DeprecationWarning(
                "except_when_invalid is deprecated. Set it to the default `True` to avoid this warning."
            )

        # if no special conditions on start and end positions
        if (allowed_start, allowed_end, deadend_start, deadend_end) == (
            None,
            None,
            False,
            False,
        ):
            positions = connected_component[
                np.random.choice(
                    len(connected_component),
                    size=2,
                    replace=False,
                )
            ]

            return self.find_shortest_path(positions[0], positions[1])

        # handle special conditions
        connected_component_set: set[CoordTup] = set(map(tuple, connected_component))
        # copy connected component set
        allowed_start_set: set[CoordTup] = connected_component_set.copy()
        allowed_end_set: set[CoordTup] = connected_component_set.copy()

        # filter by explicitly allowed start and end positions
        if allowed_start is not None:
            allowed_start_set = set(map(tuple, allowed_start)) & connected_component_set

        if allowed_end is not None:
            allowed_end_set = set(map(tuple, allowed_end)) & connected_component_set

        # filter by forcing deadends
        if deadend_start:
            allowed_start_set = set(
                filter(
                    lambda x: len(self.get_coord_neighbors(x)) == 1, allowed_start_set
                )
            )

        if deadend_end:
            allowed_end_set = set(
                filter(lambda x: len(self.get_coord_neighbors(x)) == 1, allowed_end_set)
            )

        # check we have valid positions
        if len(allowed_start_set) == 0 or len(allowed_end_set) == 0:
            raise ValueError("no valid start or end positions found")

        # randomly select start and end positions
        start_pos: CoordTup = tuple(
            list(allowed_start_set)[np.random.randint(0, len(allowed_start_set))]
        )
        end_pos: CoordTup = tuple(
            list(allowed_end_set)[np.random.randint(0, len(allowed_end_set))]
        )

        return self.find_shortest_path(start_pos, end_pos)

    # ============================================================
    # to and from adjacency list
    # ============================================================
    def as_adj_list(
        self, shuffle_d0: bool = True, shuffle_d1: bool = True
    ) -> Int8[np.ndarray, "conn start_end coord"]:
        adj_list: Int8[np.ndarray, "conn start_end coord"] = np.full(
            (self.n_connections, 2, 2),
            -1,
        )

        if shuffle_d1:
            flip_d1: Float[np.array, "conn"] = np.random.rand(self.n_connections)

        # loop over all nonzero elements of the connection list
        i: int = 0
        for d, x, y in np.ndindex(self.connection_list.shape):
            if self.connection_list[d, x, y]:
                c_start: CoordTup = (x, y)
                c_end: CoordTup = (
                    x + (1 if d == 0 else 0),
                    y + (1 if d == 1 else 0),
                )
                adj_list[i, 0] = np.array(c_start)
                adj_list[i, 1] = np.array(c_end)

                # flip if shuffling
                if shuffle_d1 and (flip_d1[i] > 0.5):
                    c_s, c_e = adj_list[i, 0].copy(), adj_list[i, 1].copy()
                    adj_list[i, 0] = c_e
                    adj_list[i, 1] = c_s

                i += 1

        if shuffle_d0:
            np.random.shuffle(adj_list)

        return adj_list

    @classmethod
    def from_adj_list(
        cls,
        adj_list: Int8[np.ndarray, "conn start_end coord"],
    ) -> "LatticeMaze":
        """create a LatticeMaze from a list of connections"""

        # Note: This has only been tested for square mazes. Might need to change some things if rectangular mazes are needed.
        grid_n: int = adj_list.max() + 1

        connection_list: ConnectionList = np.zeros(
            (2, grid_n, grid_n),
            dtype=np.bool_,
        )

        for c_start, c_end in adj_list:
            # check that exactly 1 coordinate matches
            if (c_start == c_end).sum() != 1:
                raise ValueError("invalid connection")

            # get the direction
            d: int = (c_start != c_end).argmax()

            x: int
            y: int
            # pick whichever has the lesser value in the direction `d`
            if c_start[d] < c_end[d]:
                x, y = c_start
            else:
                x, y = c_end

            connection_list[d, x, y] = True

        return LatticeMaze(
            connection_list=connection_list,
        )

    def as_adj_list_tokens(self) -> list[str | CoordTup]:
        return [
            SPECIAL_TOKENS.ADJLIST_START,
            *chain.from_iterable(
                [
                    [
                        tuple(c_s),
                        SPECIAL_TOKENS.CONNECTOR,
                        tuple(c_e),
                        SPECIAL_TOKENS.ADJACENCY_ENDLINE,
                    ]
                    for c_s, c_e in self.as_adj_list()
                ]
            ),
            SPECIAL_TOKENS.ADJLIST_END,
        ]

    def _as_coords_and_special_AOTP(self) -> list[CoordTup | str]:
        """turn the maze into adjacency list, origin, target, and solution -- keep coords as tuples"""
        output: list[str] = self.as_adj_list_tokens()
        # if getattr(self, "start_pos", None) is not None:
        if isinstance(self, TargetedLatticeMaze):
            output += self.get_start_pos_tokens()
        if isinstance(self, TargetedLatticeMaze):
            output += self.get_end_pos_tokens()
        if isinstance(self, SolvedMaze):
            output += self.get_solution_tokens()
        return output

    def as_tokens(
        self,
        maze_tokenizer: MazeTokenizer | TokenizationMode,
    ) -> list[str]:
        """serialize maze and solution to tokens"""
        if isinstance(maze_tokenizer, TokenizationMode):
            maze_tokenizer = MazeTokenizer(maze_tokenizer)
        if maze_tokenizer.is_AOTP():
            coords_raw: list[CoordTup | str] = self._as_coords_and_special_AOTP()
            coords_processed: list[str] = maze_tokenizer.coords_to_strings(
                coords=coords_raw, when_noncoord="include"
            )
            return coords_processed
        else:
            raise NotImplementedError("only AOTP tokenization is supported")

    @classmethod
    def _from_tokens_AOTP(
        cls, tokens: list[str], maze_tokenizer: MazeTokenizer
    ) -> "LatticeMaze":
        """create a LatticeMaze from a list of tokens"""

        # figure out what input format
        # ========================================
        if tokens[0] == SPECIAL_TOKENS.ADJLIST_START:
            adj_list_tokens = get_adj_list_tokens(tokens)
        else:
            # If we're not getting a "complete" tokenized maze, assume it's just a the adjacency list tokens
            adj_list_tokens = tokens
            warnings.warn(
                "Assuming input is just adjacency list tokens, no special tokens found"
            )

        # process edges for adjacency list
        # ========================================
        edges: list[list[str]] = list_split(
            adj_list_tokens,
            SPECIAL_TOKENS.ADJACENCY_ENDLINE,
        )

        coordinates: list[tuple[CoordTup, CoordTup]] = list()
        for e in edges:
            # skip last endline
            if len(e) != 0:
                # convert to coords, split start and end
                e_coords: list[str | CoordTup] = maze_tokenizer.strings_to_coords(
                    e, when_noncoord="include"
                )
                assert len(e_coords) == 3, f"invalid edge: {e = } {e_coords = }"
                assert (
                    e_coords[1] == SPECIAL_TOKENS.CONNECTOR
                ), f"invalid edge: {e = } {e_coords = }"
                coordinates.append((e_coords[0], e_coords[-1]))

        assert all(
            len(c) == 2 for c in coordinates
        ), f"invalid coordinates: {coordinates = }"
        adj_list: Int8[np.ndarray, "conn start_end coord"] = np.array(coordinates)
        assert tuple(adj_list.shape) == (
            len(coordinates),
            2,
            2,
        ), f"invalid adj_list: {adj_list.shape = } {coordinates = }"

        output_maze: LatticeMaze = cls.from_adj_list(adj_list)

        # add start and end positions
        # ========================================
        is_targeted: bool = False
        if all(
            x in tokens
            for x in (
                SPECIAL_TOKENS.ORIGIN_START,
                SPECIAL_TOKENS.ORIGIN_END,
                SPECIAL_TOKENS.TARGET_START,
                SPECIAL_TOKENS.TARGET_END,
            )
        ):
            start_pos_list: list[CoordTup] = maze_tokenizer.strings_to_coords(
                get_origin_tokens(tokens), when_noncoord="error"
            )
            end_pos_list: list[CoordTup] = maze_tokenizer.strings_to_coords(
                get_target_tokens(tokens), when_noncoord="error"
            )
            assert (
                len(start_pos_list) == 1
            ), f"invalid start_pos_list: {start_pos_list = }"
            assert len(end_pos_list) == 1, f"invalid end_pos_list: {end_pos_list = }"

            start_pos: CoordTup = start_pos_list[0]
            end_pos: CoordTup = end_pos_list[0]

            output_maze = TargetedLatticeMaze.from_lattice_maze(
                lattice_maze=output_maze,
                start_pos=start_pos,
                end_pos=end_pos,
            )

            is_targeted = True

        if all(
            x in tokens for x in (SPECIAL_TOKENS.PATH_START, SPECIAL_TOKENS.PATH_END)
        ):
            assert is_targeted, "maze must be targeted to have a solution"
            solution: list[CoordTup] = maze_tokenizer.strings_to_coords(
                get_path_tokens(tokens, trim_end=True),
                when_noncoord="error",
            )
            output_maze = SolvedMaze.from_targeted_lattice_maze(
                targeted_lattice_maze=output_maze,
                solution=solution,
            )

        return output_maze

    @classmethod
    def from_tokens(
        cls, tokens: list[str], maze_tokenizer: MazeTokenizer | TokenizationMode
    ) -> "LatticeMaze":
        if isinstance(maze_tokenizer, TokenizationMode):
            maze_tokenizer = MazeTokenizer(maze_tokenizer)

        if isinstance(tokens, str):
            tokens = tokens.split()

        if maze_tokenizer.is_AOTP():
            return cls._from_tokens_AOTP(tokens, maze_tokenizer)
        else:
            raise NotImplementedError("only AOTP tokenization is supported")

    # ============================================================
    # to and from pixels
    # ============================================================
    def _as_pixels_bw(self) -> BinaryPixelGrid:
        assert self.lattice_dim == 2, "only 2D mazes are supported"
        # Create an empty pixel grid with walls
        pixel_grid: Int[np.ndarray, "x y"] = np.full(
            (self.grid_shape[0] * 2 + 1, self.grid_shape[1] * 2 + 1),
            False,
            dtype=np.bool_,
        )

        # Set white nodes
        pixel_grid[1::2, 1::2] = True

        # Set white connections (downward)
        for i, row in enumerate(self.connection_list[0]):
            for j, connected in enumerate(row):
                if connected:
                    pixel_grid[i * 2 + 2, j * 2 + 1] = True

        # Set white connections (rightward)
        for i, row in enumerate(self.connection_list[1]):
            for j, connected in enumerate(row):
                if connected:
                    pixel_grid[i * 2 + 1, j * 2 + 2] = True

        return pixel_grid

    def as_pixels(
        self,
        show_endpoints: bool = True,
        show_solution: bool = True,
    ) -> PixelGrid:
        if show_solution and not show_endpoints:
            raise ValueError("show_solution=True requires show_endpoints=True")
        # convert original bool pixel grid to RGB
        pixel_grid_bw: BinaryPixelGrid = self._as_pixels_bw()
        pixel_grid: PixelGrid = np.full(
            (*pixel_grid_bw.shape, 3), PixelColors.WALL, dtype=np.uint8
        )
        pixel_grid[pixel_grid_bw == True] = PixelColors.OPEN

        if self.__class__ == LatticeMaze:
            return pixel_grid

        # set endpoints for TargetedLatticeMaze
        if self.__class__ == TargetedLatticeMaze:
            if show_endpoints:
                pixel_grid[self.start_pos[0] * 2 + 1, self.start_pos[1] * 2 + 1] = (
                    PixelColors.START
                )
                pixel_grid[self.end_pos[0] * 2 + 1, self.end_pos[1] * 2 + 1] = (
                    PixelColors.END
                )
            return pixel_grid

        # set solution
        if show_solution:
            for coord in self.solution:
                pixel_grid[coord[0] * 2 + 1, coord[1] * 2 + 1] = PixelColors.PATH

            # set pixels between coords
            for index, coord in enumerate(self.solution[:-1]):
                next_coord = self.solution[index + 1]
                # check they are adjacent using norm
                assert (
                    np.linalg.norm(np.array(coord) - np.array(next_coord)) == 1
                ), f"Coords {coord} and {next_coord} are not adjacent"
                # set pixel between them
                pixel_grid[
                    coord[0] * 2 + 1 + next_coord[0] - coord[0],
                    coord[1] * 2 + 1 + next_coord[1] - coord[1],
                ] = PixelColors.PATH

            # set endpoints (again, since path would overwrite them)
            pixel_grid[self.start_pos[0] * 2 + 1, self.start_pos[1] * 2 + 1] = (
                PixelColors.START
            )
            pixel_grid[self.end_pos[0] * 2 + 1, self.end_pos[1] * 2 + 1] = (
                PixelColors.END
            )

        return pixel_grid

    @classmethod
    def _from_pixel_grid_bw(
        cls, pixel_grid: BinaryPixelGrid
    ) -> tuple[ConnectionList, tuple[int, int]]:
        grid_shape: tuple[int, int] = (
            pixel_grid.shape[0] // 2,
            pixel_grid.shape[1] // 2,
        )
        connection_list: ConnectionList = np.zeros((2, *grid_shape), dtype=np.bool_)

        # Extract downward connections
        connection_list[0] = pixel_grid[2::2, 1::2]

        # Extract rightward connections
        connection_list[1] = pixel_grid[1::2, 2::2]

        return connection_list, grid_shape

    @classmethod
    def _from_pixel_grid_with_positions(
        cls,
        pixel_grid: PixelGrid | BinaryPixelGrid,
        marked_positions: dict[str, RGB],
    ) -> tuple[ConnectionList, tuple[int, int], dict[str, CoordArray]]:
        # Convert RGB pixel grid to Bool pixel grid
        pixel_grid_bw: BinaryPixelGrid = ~np.all(
            pixel_grid == PixelColors.WALL, axis=-1
        )
        connection_list: ConnectionList
        grid_shape: tuple[int, int]
        connection_list, grid_shape = cls._from_pixel_grid_bw(pixel_grid_bw)

        # Find any marked positions
        out_positions: dict[str, CoordArray] = dict()
        for key, color in marked_positions.items():
            pos_temp: Int[np.ndarray, "x y"] = np.argwhere(
                np.all(pixel_grid == color, axis=-1)
            )
            pos_save: list[CoordTup] = list()
            for pos in pos_temp:
                # if it is a coordinate and not connection (transform position, %2==1)
                if pos[0] % 2 == 1 and pos[1] % 2 == 1:
                    pos_save.append((pos[0] // 2, pos[1] // 2))

            out_positions[key] = np.array(pos_save)

        return connection_list, grid_shape, out_positions

    @classmethod
    def from_pixels(
        cls,
        pixel_grid: PixelGrid,
    ) -> "LatticeMaze":
        connection_list: ConnectionList
        grid_shape: tuple[int, int]

        # if a binary pixel grid, return regular LaticeMaze
        if len(pixel_grid.shape) == 2:
            connection_list, grid_shape = cls._from_pixel_grid_bw(pixel_grid)
            return LatticeMaze(connection_list=connection_list)

        # otherwise, detect and check it's valid
        cls_detected: typing.Type[LatticeMaze] = detect_pixels_type(pixel_grid)
        if not cls in cls_detected.__mro__:
            raise ValueError(
                f"Pixel grid cannot be cast to {cls.__name__}, detected type {cls_detected.__name__}"
            )

        (
            connection_list,
            grid_shape,
            marked_pos,
        ) = cls._from_pixel_grid_with_positions(
            pixel_grid=pixel_grid,
            marked_positions=dict(
                start=PixelColors.START, end=PixelColors.END, solution=PixelColors.PATH
            ),
        )
        # if we wanted a LatticeMaze, return it
        if cls == LatticeMaze:
            return LatticeMaze(connection_list=connection_list)

        # otherwise, keep going
        temp_maze: LatticeMaze = LatticeMaze(connection_list=connection_list)

        # start and end pos
        start_pos_arr, end_pos_arr = marked_pos["start"], marked_pos["end"]
        assert start_pos_arr.shape == (
            1,
            2,
        ), f"start_pos_arr {start_pos_arr} has shape {start_pos_arr.shape}, expected shape (1, 2) -- a single coordinate"
        assert end_pos_arr.shape == (
            1,
            2,
        ), f"end_pos_arr {end_pos_arr} has shape {end_pos_arr.shape}, expected shape (1, 2) -- a single coordinate"

        start_pos: Coord = start_pos_arr[0]
        end_pos: Coord = end_pos_arr[0]

        # return a TargetedLatticeMaze if that's what we wanted
        if cls == TargetedLatticeMaze:
            return TargetedLatticeMaze(
                connection_list=connection_list,
                start_pos=start_pos,
                end_pos=end_pos,
            )

        # raw solution, only contains path elements and not start or end
        solution_raw: CoordArray = marked_pos["solution"]
        if len(solution_raw.shape) == 2:
            assert (
                solution_raw.shape[1] == 2
            ), f"solution {solution_raw} has shape {solution_raw.shape}, expected shape (n, 2)"
        elif solution_raw.shape == (0,):
            # the solution and end should be immediately adjacent
            assert (
                np.sum(np.abs(start_pos - end_pos)) == 1
            ), f"start_pos {start_pos} and end_pos {end_pos} are not adjacent, but no solution was given"

        # order the solution, by creating a list from the start to the end
        # add end pos, since we will iterate over all these starting from the start pos
        solution_raw_list: list[CoordTup] = [tuple(c) for c in solution_raw] + [
            tuple(end_pos)
        ]
        # solution starts with start point
        solution: list[CoordTup] = [tuple(start_pos)]
        while solution[-1] != tuple(end_pos):
            # use `get_coord_neighbors` to find connected neighbors
            neighbors: CoordArray = temp_maze.get_coord_neighbors(solution[-1])
            # TODO: make this less ugly
            assert (len(neighbors.shape) == 2) and (
                neighbors.shape[1] == 2
            ), f"neighbors {neighbors} has shape {neighbors.shape}, expected shape (n, 2)\n{neighbors = }\n{solution = }\n{solution_raw = }\n{temp_maze.as_ascii()}"
            # neighbors = neighbors[:, [1, 0]]
            # filter out neighbors that are not in the raw solution
            neighbors_filtered: CoordArray = np.array(
                [
                    coord
                    for coord in neighbors
                    if (
                        tuple(coord) in solution_raw_list
                        and not tuple(coord) in solution
                    )
                ]
            )
            # assert only one element is left, and then add it to the solution
            assert neighbors_filtered.shape == (
                1,
                2,
            ), f"neighbors_filtered has shape {neighbors_filtered.shape}, expected shape (1, 2)\n{neighbors = }\n{neighbors_filtered = }\n{solution = }\n{solution_raw_list = }\n{temp_maze.as_ascii()}"
            solution.append(tuple(neighbors_filtered[0]))

        # assert the solution is complete
        assert solution[0] == tuple(
            start_pos
        ), f"solution {solution} does not start at start_pos {start_pos}"
        assert solution[-1] == tuple(
            end_pos
        ), f"solution {solution} does not end at end_pos {end_pos}"

        return cls(
            connection_list=np.array(connection_list),
            solution=np.array(solution),
        )

    # ============================================================
    # to and from ASCII
    # ============================================================
    def _as_ascii_grid(self) -> Shaped[np.ndarray, "x y"]:
        # Get the pixel grid using to_pixels().
        pixel_grid: Bool[np.ndarray, "x y"] = self._as_pixels_bw()

        # Replace pixel values with ASCII characters.
        ascii_grid: Shaped[np.ndarray, "x y"] = np.full(
            pixel_grid.shape, AsciiChars.WALL, dtype=str
        )
        ascii_grid[pixel_grid == True] = AsciiChars.OPEN

        return ascii_grid

    def as_ascii(
        self,
        show_endpoints: bool = True,
        show_solution: bool = True,
    ) -> str:
        """return an ASCII grid of the maze"""
        ascii_grid: Shaped[np.ndarray, "x y"] = self._as_ascii_grid()
        pixel_grid: PixelGrid = self.as_pixels(
            show_endpoints=show_endpoints, show_solution=show_solution
        )

        chars_replace: tuple = tuple()
        if show_endpoints:
            chars_replace += (AsciiChars.START, AsciiChars.END)
        if show_solution:
            chars_replace += (AsciiChars.PATH,)

        for ascii_char, pixel_color in ASCII_PIXEL_PAIRINGS.items():
            if ascii_char in chars_replace:
                ascii_grid[(pixel_grid == pixel_color).all(axis=-1)] = ascii_char

        return "\n".join("".join(row) for row in ascii_grid)

    @classmethod
    def from_ascii(cls, ascii_str: str) -> "LatticeMaze":
        lines: list[str] = ascii_str.strip().split("\n")
        lines = [line.strip() for line in lines]
        ascii_grid: Shaped[np.ndarray, "x y"] = np.array(
            [list(line) for line in lines], dtype=str
        )
        pixel_grid: PixelGrid = np.zeros((*ascii_grid.shape, 3), dtype=np.uint8)

        for ascii_char, pixel_color in ASCII_PIXEL_PAIRINGS.items():
            pixel_grid[ascii_grid == ascii_char] = pixel_color

        return cls.from_pixels(pixel_grid)


@serializable_dataclass(frozen=True, kw_only=True)
class TargetedLatticeMaze(LatticeMaze):
    """A LatticeMaze with a start and end position"""

    # this jank is so that SolvedMaze can inherit from this class without needing arguments for start_pos and end_pos
    start_pos: Coord
    end_pos: Coord

    def __post_init__(self) -> None:
        # make things numpy arrays (very jank to override frozen dataclass)
        self.__dict__["start_pos"] = np.array(self.start_pos)
        self.__dict__["end_pos"] = np.array(self.end_pos)
        assert self.start_pos is not None
        assert self.end_pos is not None
        # check that start and end are in bounds
        if (
            self.start_pos[0] >= self.grid_shape[0]
            or self.start_pos[1] >= self.grid_shape[1]
        ):
            raise ValueError(
                f"start_pos {self.start_pos} is out of bounds for grid shape {self.grid_shape}"
            )
        if (
            self.end_pos[0] >= self.grid_shape[0]
            or self.end_pos[1] >= self.grid_shape[1]
        ):
            raise ValueError(
                f"end_pos {self.end_pos} is out of bounds for grid shape {self.grid_shape}"
            )

    def get_start_pos_tokens(self) -> list[str | CoordTup]:
        return [
            SPECIAL_TOKENS.ORIGIN_START,
            tuple(self.start_pos),
            SPECIAL_TOKENS.ORIGIN_END,
        ]

    def get_end_pos_tokens(self) -> list[str | CoordTup]:
        return [
            SPECIAL_TOKENS.TARGET_START,
            tuple(self.end_pos),
            SPECIAL_TOKENS.TARGET_END,
        ]

    @classmethod
    def from_lattice_maze(
        cls,
        lattice_maze: LatticeMaze,
        start_pos: Coord,
        end_pos: Coord,
    ) -> "TargetedLatticeMaze":
        return cls(
            connection_list=lattice_maze.connection_list,
            start_pos=start_pos,
            end_pos=end_pos,
            generation_meta=lattice_maze.generation_meta,
        )


@serializable_dataclass(frozen=True, kw_only=True)
class SolvedMaze(TargetedLatticeMaze):
    """Stores a maze and a solution"""

    solution: CoordArray

    def __init__(
        self,
        connection_list: ConnectionList,
        solution: CoordArray,
        generation_meta: dict | None = None,
        start_pos: Coord | None = None,
        end_pos: Coord | None = None,
        allow_invalid: bool = False,
    ) -> None:
        # figure out the solution
        solution_valid: bool = False
        if solution is not None:
            solution = np.array(solution)
            # note that a path length of 1 here is valid, since the start and end pos could be the same
            if (solution.shape[0] > 0) and (solution.shape[1] == 2):
                solution_valid = True

        if not solution_valid and not allow_invalid:
            raise ValueError(
                f"invalid solution: {solution.shape = } {solution = } {solution_valid = } {allow_invalid = }",
                f"{connection_list = }",
            )

        # init the TargetedLatticeMaze
        super().__init__(
            connection_list=connection_list,
            generation_meta=generation_meta,
            start_pos=np.array(solution[0]) if solution_valid else None,
            end_pos=np.array(solution[-1]) if solution_valid else None,
        )

        self.__dict__["solution"] = solution

        # adjust the endpoints
        if not allow_invalid:
            if start_pos is not None:
                assert np.array_equal(
                    np.array(start_pos), self.start_pos
                ), f"when trying to create a SolvedMaze, the given start_pos does not match the one in the solution: given={start_pos}, solution={self.start_pos}"
            if end_pos is not None:
                assert np.array_equal(
                    np.array(end_pos), self.end_pos
                ), f"when trying to create a SolvedMaze, the given end_pos does not match the one in the solution: given={end_pos}, solution={self.end_pos}"
            # TODO: assert the path does not backtrack, walk through walls, etc?

    def __hash__(self) -> int:
        return hash((self.connection_list.tobytes(), self.solution.tobytes()))

    def get_solution_tokens(self) -> list[str | CoordTup]:
        return [
            SPECIAL_TOKENS.PATH_START,
            *[tuple(c) for c in self.solution],
            SPECIAL_TOKENS.PATH_END,
        ]

    # for backwards compatibility
    @property
    def maze(self) -> LatticeMaze:
        warnings.warn(
            "maze is deprecated, SolvedMaze now inherits from LatticeMaze.",
            DeprecationWarning,
        )
        return LatticeMaze(connection_list=self.connection_list)

    @classmethod
    def from_lattice_maze(
        cls, lattice_maze: LatticeMaze, solution: list[CoordTup]
    ) -> "SolvedMaze":
        return cls(
            connection_list=lattice_maze.connection_list,
            solution=solution,
            generation_meta=lattice_maze.generation_meta,
        )

    @classmethod
    def from_targeted_lattice_maze(
        cls,
        targeted_lattice_maze: TargetedLatticeMaze,
        solution: list[CoordTup] | None = None,
    ) -> "SolvedMaze":
        """solves the given targeted lattice maze and returns a SolvedMaze"""
        if solution is None:
            solution = targeted_lattice_maze.find_shortest_path(
                targeted_lattice_maze.start_pos,
                targeted_lattice_maze.end_pos,
            )
        return cls(
            connection_list=targeted_lattice_maze.connection_list,
            solution=np.array(solution),
            generation_meta=targeted_lattice_maze.generation_meta,
        )

    def get_solution_forking_points(self) -> tuple[list[int], CoordArray]:
        """coordinates and their indicies from the solution where a fork is present

        - if the start point is not a dead end, this counts as a fork
        - if the end point is not a dead end, this counts as a fork
        """
        output_idxs: list[int] = list()
        output_coords: list[CoordTup] = list()

        for idx, coord in enumerate(self.solution):
            # more than one choice for first coord, or more than 2 for any other
            # since the previous coord doesn't count as a choice
            is_endpoint: bool = idx == 0 or idx == self.solution.shape[0] - 1
            theshold: int = 1 if is_endpoint else 2
            if self.get_coord_neighbors(coord).shape[0] > theshold:
                output_idxs.append(idx)
                output_coords.append(coord)

        return output_idxs, np.array(output_coords)

    def get_solution_path_following_points(self) -> tuple[list[int], CoordArray]:
        """coordinates from the solution where there is only a single (non-backtracking) point to move to

        returns the complement of `get_solution_forking_points` from the path"""
        forks_idxs, _ = self.get_solution_forking_points()
        return (
            np.delete(np.arange(self.solution.shape[0]), forks_idxs, axis=0),
            np.delete(self.solution, forks_idxs, axis=0),
        )


def detect_pixels_type(data: PixelGrid) -> typing.Type[LatticeMaze]:
    """Detects the type of pixels data by checking for the presence of start and end pixels"""
    if color_in_pixel_grid(data, PixelColors.START) or color_in_pixel_grid(
        data, PixelColors.END
    ):
        if color_in_pixel_grid(data, PixelColors.PATH):
            return SolvedMaze
        else:
            return TargetedLatticeMaze
    else:
        return LatticeMaze


def _remove_isolated_cells(
    image: Int[np.ndarray, "RGB x y"]
) -> Int[np.ndarray, "RGB x y"]:
    """
    Removes isolated cells from an image. An isolated cell is a cell that is surrounded by walls on all sides.
    """
    # Create a binary mask where True represents walls
    wall_mask = np.all(image == PixelColors.WALL, axis=-1)

    # Pad the wall mask to handle edge cases
    padded_wall_mask = np.pad(
        wall_mask, ((1, 1), (1, 1)), mode="constant", constant_values=True
    )

    # Check neighbors in all four directions
    isolated_mask = (
        padded_wall_mask[1:-1, 2:]  # right
        & padded_wall_mask[1:-1, :-2]  # left
        & padded_wall_mask[2:, 1:-1]  # down
        & padded_wall_mask[:-2, 1:-1]  # up
    )

    # Combine with non-wall mask to only affect open cells
    non_wall_mask = ~wall_mask
    isolated_mask = isolated_mask & non_wall_mask

    # Create the output image
    output_image = image.copy()
    output_image[isolated_mask] = PixelColors.WALL

    return output_image


_RIC_PADS: dict = {
    "left": ((1, 0), (0, 0)),
    "right": ((0, 1), (0, 0)),
    "up": ((0, 0), (1, 0)),
    "down": ((0, 0), (0, 1)),
}

# Define slices for each direction
_RIC_SLICES: dict = {
    "left": (slice(1, None), slice(None, None)),
    "right": (slice(None, -1), slice(None, None)),
    "up": (slice(None, None), slice(1, None)),
    "down": (slice(None, None), slice(None, -1)),
}


def _remove_isolated_cells_old(
    image: Int[np.ndarray, "RGB x y"]
) -> Int[np.ndarray, "RGB x y"]:
    """
    Removes isolated cells from an image. An isolated cell is a cell that is surrounded by walls on all sides.
    """
    warnings.warn("this functin doesn't work and I have no idea why!!!")
    masks: dict[str, np.ndarray] = {
        d: np.all(
            np.pad(
                image[_RIC_SLICES[d][0], _RIC_SLICES[d][1], :] == PixelColors.WALL,
                np.array((*_RIC_PADS[d], (0, 0)), dtype=np.int8),
                mode="constant",
                constant_values=True,
            ),
            axis=2,
        )
        for d in _RIC_SLICES.keys()
    }

    # Create a mask for non-wall cells
    mask_non_wall = np.all(image != PixelColors.WALL, axis=2)

    # print(f"{mask_non_wall.shape = }")
    # print(f"{ {k: masks[k].shape for k in masks.keys()} = }")

    # print(f"{mask_non_wall = }")
    # print(f"{masks['down'] = }")

    # Combine the masks
    mask = mask_non_wall & masks["left"] & masks["right"] & masks["up"] & masks["down"]

    # Apply the mask
    output_image = np.where(
        np.stack([mask] * 3, axis=-1),
        PixelColors.WALL,
        image,
    )

    return output_image



from maze_dataset.maze.lattice_maze import (
    AsciiChars,
    ConnectionList,
    Coord,
    CoordArray,
    LatticeMaze,
    PixelColors,
    SolvedMaze,
    TargetedLatticeMaze,
)

__all__ = [
    "AsciiChars",
    "ConnectionList",
    "Coord",
    "CoordArray",
    "LatticeMaze",
    "PixelColors",
    "SolvedMaze",
    "TargetedLatticeMaze",
]



"""a whole bunch of utilities for tokenization"""

import warnings

from maze_dataset.constants import SPECIAL_TOKENS
from maze_dataset.tokenization.maze_tokenizer import TokenizationMode, is_UT

# filtering things from a prompt or generated text
# ==================================================


def remove_padding_from_token_str(token_str: str) -> str:
    token_str = token_str.replace(f"{SPECIAL_TOKENS.PADDING} ", "")
    token_str = token_str.replace(f"{SPECIAL_TOKENS.PADDING}", "")
    return token_str


def tokens_between(
    tokens: list[str],
    start_value: str,
    end_value: str,
    include_start: bool = False,
    include_end: bool = False,
    except_when_tokens_not_unique: bool = False,
) -> list[str]:
    if start_value == end_value:
        raise ValueError(
            f"start_value and end_value cannot be the same: {start_value = } {end_value = }"
        )
    if except_when_tokens_not_unique:
        if (tokens.count(start_value) != 1) or (tokens.count(end_value) != 1):
            raise ValueError(
                "start_value or end_value is not unique in the input tokens:",
                f"{tokens.count(start_value) = } {tokens.count(end_value) = }"
                f"{start_value = } {end_value = }",
                f"{tokens = }",
            )
    else:
        if (tokens.count(start_value) < 1) or (tokens.count(end_value) < 1):
            raise ValueError(
                "start_value or end_value is not present in the input tokens:",
                f"{tokens.count(start_value) = } {tokens.count(end_value) = }",
                f"{start_value = } {end_value = }",
                f"{tokens = }",
            )

    start_idx: int = tokens.index(start_value) + int(not include_start)
    end_idx: int = tokens.index(end_value) + int(include_end)

    assert start_idx < end_idx, "Start must come before end"

    return tokens[start_idx:end_idx]


def get_adj_list_tokens(tokens: list[str]) -> list[str]:
    return tokens_between(
        tokens, SPECIAL_TOKENS.ADJLIST_START, SPECIAL_TOKENS.ADJLIST_END
    )


def get_path_tokens(tokens: list[str], trim_end: bool = False) -> list[str]:
    """The path is considered everything from the first path coord to the path_end token, if it exists."""
    if SPECIAL_TOKENS.PATH_START not in tokens:
        raise ValueError(
            f"Path start token {SPECIAL_TOKENS.PATH_START} not found in tokens:\n{tokens}"
        )
    start_idx: int = tokens.index(SPECIAL_TOKENS.PATH_START) + int(trim_end)
    end_idx: int | None = None
    if trim_end and (SPECIAL_TOKENS.PATH_END in tokens):
        end_idx = tokens.index(SPECIAL_TOKENS.PATH_END)
    return tokens[start_idx:end_idx]


def get_context_tokens(tokens: list[str]) -> list[str]:
    return tokens_between(
        tokens,
        SPECIAL_TOKENS.ADJLIST_START,
        SPECIAL_TOKENS.PATH_START,
        include_start=True,
        include_end=True,
    )


def get_origin_tokens(tokens: list[str]) -> list[str]:
    return tokens_between(
        tokens,
        SPECIAL_TOKENS.ORIGIN_START,
        SPECIAL_TOKENS.ORIGIN_END,
        include_start=False,
        include_end=False,
    )


def get_target_tokens(tokens: list[str]) -> list[str]:
    return tokens_between(
        tokens,
        SPECIAL_TOKENS.TARGET_START,
        SPECIAL_TOKENS.TARGET_END,
        include_start=False,
        include_end=False,
    )


def get_tokens_up_to_path_start(
    tokens: list[str],
    include_start_coord: bool = True,
    tokenization_mode: TokenizationMode = TokenizationMode.AOTP_UT_uniform,
) -> list[str]:
    warnings.warn(
        "`get_tokens_up_to_path_start` assumes a unique token (UT) type tokenizer when `include_start_coord=True`. "
        "This method is deprecated for a tokenizer-agnostic function in a future release.",
        PendingDeprecationWarning,
    )
    path_start_idx: int = tokens.index(SPECIAL_TOKENS.PATH_START) + 1
    if include_start_coord:
        if is_UT(tokenization_mode):
            return tokens[: path_start_idx + 1]
        elif tokenization_mode == TokenizationMode.AOTP_CTT_indexed:
            return tokens[: path_start_idx + 5]
        else:
            raise ValueError(f"Invalid tokenization mode: {tokenization_mode}")
    else:
        return tokens[:path_start_idx]



from maze_dataset.tokenization.maze_tokenizer import MazeTokenizer, TokenizationMode
from maze_dataset.tokenization.token_utils import (
    get_adj_list_tokens,
    get_context_tokens,
    get_origin_tokens,
    get_path_tokens,
    get_target_tokens,
    get_tokens_up_to_path_start,
    tokens_between,
)
from maze_dataset.tokenization.util import coord_str_to_tuple

__all__ = [
    "MazeTokenizer",
    "TokenizationMode",
    "coord_str_to_tuple",
    "get_adj_list_tokens",
    "get_context_tokens",
    "get_origin_tokens",
    "get_path_tokens",
    "get_target_tokens",
    "get_tokens_up_to_path_start",
    "tokens_between",
]



from maze_dataset.generation.generators import (
    GENERATORS_MAP,
    LatticeMazeGenerators,
    get_maze_with_solution,
)

__all__ = [
    "GENERATORS_MAP",
    "LatticeMazeGenerators",
    "get_maze_with_solution",
]



"""utilities required by MazeTokenizer"""

import re
import typing
from typing import Callable

import numpy as np
from muutils.misc import list_join

from maze_dataset.constants import CoordTup
from maze_dataset.utils import WhenMissing

# coordinate to strings
# ==================================================


def _coord_to_strings_UT(coord: typing.Sequence[int]) -> list[str]:
    """convert a coordinate to a string: `(i,j)`->"(i,j)"
    always returns a list of length 1"""
    return [f"({','.join(str(c) for c in coord)})"]


def _coord_to_strings_indexed(coord: typing.Sequence[int]) -> list[str]:
    """convert a coordinate to a list of indexed strings: `(i,j)`->"(", "i", ",", "j", ")" """
    return [
        "(",
        *list_join([str(c) for c in coord], lambda: ","),
        ")",
    ]


# string to coordinate representation
# ==================================================


def str_is_coord(coord_str: str, allow_whitespace: bool = True) -> bool:
    """return True if the string represents a coordinate, False otherwise"""

    strip_func: Callable[[str], str] = lambda x: x.strip() if allow_whitespace else x

    coord_str = strip_func(coord_str)

    return all(
        [
            coord_str.startswith("("),
            coord_str.endswith(")"),
            "," in coord_str,
            all(
                [
                    strip_func(x).isdigit()
                    for x in strip_func(coord_str.lstrip("(").rstrip(")")).split(",")
                ]
            ),
        ]
    )


def coord_str_to_tuple(
    coord_str: str, allow_whitespace: bool = True
) -> tuple[int, ...]:
    """convert a coordinate string to a tuple"""
    strip_func: Callable[[str], str] = lambda x: x.strip() if allow_whitespace else x
    coord_str = strip_func(coord_str)
    stripped: str = strip_func(coord_str.lstrip("(").rstrip(")"))
    return tuple(int(strip_func(x)) for x in stripped.split(","))


def coord_str_to_coord_np(coord_str: str, allow_whitespace: bool = True) -> np.ndarray:
    """convert a coordinate string to a numpy array"""
    return np.array(coord_str_to_tuple(coord_str, allow_whitespace=allow_whitespace))


def coord_str_to_tuple_noneable(coord_str: str) -> CoordTup | None:
    """convert a coordinate string to a tuple, or None if the string is not a coordinate string"""
    if not str_is_coord(coord_str):
        return None
    return coord_str_to_tuple(coord_str)


def coords_string_split_UT(coords: str) -> list[str]:
    """Splits a string of tokens into a list containing the UT tokens for each coordinate.

    Not capable of producing indexed tokens ("(", "1", ",", "2", ")"), only unique tokens ("(1,2)").
    Non-whitespace portions of the input string not matched are preserved in the same list:
    "(1,2) <SPECIAL_TOKEN> (5,6)" -> ["(1,2)", "<SPECIAL_TOKEN>", "(5,6)"]
    """
    # ty gpt4
    return re.findall(r"\([^)]*\)|\S+", coords)


# back and forth in wrapped form
# ==================================================
def strings_to_coords(
    text: str | list[str],
    when_noncoord: WhenMissing = "skip",
) -> list[str | CoordTup]:
    """converts a list of tokens to a list of coordinates

    returns list[CoordTup] if `when_noncoord` is "skip" or "error"
    returns list[str | CoordTup] if `when_noncoord` is "include"
    """
    tokens_joined: str = text if isinstance(text, str) else " ".join(text)
    tokens_processed: list[str] = coords_string_split_UT(tokens_joined)
    result: list[str] = list()
    for token in tokens_processed:
        coord: CoordTup | None = coord_str_to_tuple_noneable(token)
        if coord is None:
            if when_noncoord == "skip":
                continue
            elif when_noncoord == "error":
                raise ValueError(
                    f"Invalid non-coordinate token '{token}' in text: '{text}'"
                )
            elif when_noncoord == "include":
                result.append(token)
            else:
                raise ValueError(f"Invalid when_noncoord value '{when_noncoord}'")
        else:
            result.append(coord)
    return result


def coords_to_strings(
    coords: list[str | CoordTup],
    coord_to_strings_func: Callable[[CoordTup], list[str]],
    when_noncoord: WhenMissing = "skip",
) -> list[str]:
    """converts a list of coordinates to a list of strings (tokens)

    expects list[CoordTup] if `when_noncoord` is "error"
    expects list[str | CoordTup] if `when_noncoord` is "include" or "skip"
    """
    result: list[str] = list()
    for coord in coords:
        if isinstance(coord, str):
            if when_noncoord == "skip":
                continue
            elif when_noncoord == "error":
                raise ValueError(
                    f"Invalid non-coordinate '{coord}' in list of coords: '{coords}'"
                )
            elif when_noncoord == "include":
                result.append(coord)
            else:
                raise ValueError(f"Invalid when_noncoord value '{when_noncoord}'")
        else:
            result.extend(coord_to_strings_func(coord))
    return result



from maze_dataset.dataset.collected_dataset import (
    MazeDatasetCollection,
    MazeDatasetCollectionConfig,
)
from maze_dataset.dataset.maze_dataset import MazeDataset, MazeDatasetConfig

__all__ = [
    "MazeDataset",
    "MazeDatasetConfig",
    "MazeDatasetCollection",
    "MazeDatasetCollectionConfig",
]



import html
from typing import Literal, Sequence

import matplotlib
import numpy as np
from IPython.display import HTML, display
from jaxtyping import UInt8

from maze_dataset.constants import SPECIAL_TOKENS
from maze_dataset.tokenization import tokens_between

RGBArray = UInt8[np.ndarray, "n 3"]

FormatType = Literal["html", "latex", "terminal", None]

TEMPLATES: dict[FormatType, str] = {
    "html": '<span style="color: black; background-color: rgb({clr})">&nbsp{tok}&nbsp</span>',
    "latex": "\\colorbox[RGB]{{ {clr} }}{{ \\texttt{{ {tok} }} }}",
    "terminal": "\033[30m\033[48;2;{clr}m{tok}\033[0m",
}

_COLOR_JOIN: dict[FormatType, str] = {
    "html": ",",
    "latex": ",",
    "terminal": ";",
}


def _escape_tok(
    tok: str,
    fmt: FormatType,
) -> str:
    if fmt == "html":
        return html.escape(tok)
    elif fmt == "latex":
        return tok.replace("_", "\\_").replace("#", "\\#")
    elif fmt == "terminal":
        return tok
    else:
        raise ValueError(f"Unexpected format: {fmt}")


def color_tokens_rgb(
    tokens: list,
    colors: Sequence[Sequence[int]],
    fmt: FormatType = "html",
    template: str | None = None,
    clr_join: str | None = None,
) -> str:
    """tokens will not be escaped if `fmt` is None"""
    # process format
    if fmt is None:
        assert template is not None
        assert clr_join is not None
    else:
        assert template is None
        assert clr_join is None
        template = TEMPLATES[fmt]
        clr_join = _COLOR_JOIN[fmt]

    # put everything together
    output = [
        template.format(
            clr=clr_join.join(map(str, map(int, clr))),
            tok=_escape_tok(tok, fmt),
        )
        for tok, clr in zip(tokens, colors)
    ]

    return " ".join(output)


def color_tokens_cmap(
    tokens: list[str],
    weights: Sequence[float],
    cmap: str | matplotlib.colors.Colormap = "Blues",
    fmt: FormatType = "html",
    template: str | None = None,
    labels: bool = False,
):
    assert len(tokens) == len(weights), f"{len(tokens)} != {len(weights)}"
    weights = np.array(weights)
    # normalize weights to [0, 1]
    weights_norm = matplotlib.colors.Normalize()(weights)

    if isinstance(cmap, str):
        cmap = matplotlib.colormaps.get_cmap(cmap)

    colors: RGBArray = cmap(weights_norm)[:, :3] * 255

    output: str = color_tokens_rgb(
        tokens=tokens,
        colors=colors,
        fmt=fmt,
        template=template,
    )

    if labels:
        if fmt != "terminal":
            raise NotImplementedError("labels only supported for terminal")
        # align labels with the tokens
        output += "\n"
        for tok, weight in zip(tokens, weights):
            # 2 decimal points, left-aligned and trailing spaces to match token length
            weight_str: str = f"{weight:.1f}"
            # omit if longer than token
            if len(weight_str) > len(tok):
                weight_str = " " * len(tok)
            else:
                weight_str = weight_str.ljust(len(tok))
            output += f"{weight_str} "

    return output


# colors roughly made to be similar to visual representation
_MAZE_TOKENS_DEFAULT_COLORS: dict[tuple[str, str], tuple[int, int, int]] = {
    (SPECIAL_TOKENS.ADJLIST_START, SPECIAL_TOKENS.ADJLIST_END): (
        217,
        210,
        233,
    ),  # purple
    (SPECIAL_TOKENS.ORIGIN_START, SPECIAL_TOKENS.ORIGIN_END): (217, 234, 211),  # green
    (SPECIAL_TOKENS.TARGET_START, SPECIAL_TOKENS.TARGET_END): (234, 209, 220),  # red
    (SPECIAL_TOKENS.PATH_START, SPECIAL_TOKENS.PATH_END): (207, 226, 243),  # blue
}


def color_maze_tokens_AOTP(
    tokens: list[str],
    fmt: FormatType = "html",
    template: str | None = None,
) -> str:
    output: list[str] = [
        " ".join(
            tokens_between(
                tokens, start_tok, end_tok, include_start=True, include_end=True
            )
        )
        for start_tok, end_tok in _MAZE_TOKENS_DEFAULT_COLORS.keys()
    ]

    colors: RGBArray = np.array(
        list(_MAZE_TOKENS_DEFAULT_COLORS.values()), dtype=np.uint8
    )

    return color_tokens_rgb(
        tokens=output,
        colors=colors,
        fmt=fmt,
        template=template,
    )


def display_html(html: str):
    display(HTML(html))


def display_color_tokens_rgb(
    tokens: list[str],
    colors: RGBArray,
) -> None:
    html: str = color_tokens_rgb(tokens, colors, fmt="html")
    display_html(html)


def display_color_tokens_cmap(
    tokens: list[str],
    weights: Sequence[float],
    cmap: str | matplotlib.colors.Colormap = "Blues",
) -> None:
    html: str = color_tokens_cmap(tokens, weights, cmap)
    display_html(html)


def display_color_maze_tokens_AOTP(
    tokens: list[str],
) -> None:
    html: str = color_maze_tokens_AOTP(tokens)
    display_html(html)



import itertools
import json
import typing
from functools import cached_property

import numpy as np
from jaxtyping import Int
from muutils.json_serialize import (
    JSONitem,
    json_serialize,
    serializable_dataclass,
    serializable_field,
)
from muutils.misc import sanitize_fname, shorten_numerical_to_str, stable_hash
from zanj.loading import LoaderHandler, load_item_recursive, register_loader_handler

from maze_dataset.constants import Coord, CoordTup
from maze_dataset.dataset.dataset import GPTDataset, GPTDatasetConfig
from maze_dataset.dataset.maze_dataset import MazeDataset, MazeDatasetConfig
from maze_dataset.maze import LatticeMaze


@serializable_dataclass(kw_only=True)
class MazeDatasetCollectionConfig(GPTDatasetConfig):
    """maze dataset collection configuration, including tokenizers and shuffle"""

    maze_dataset_configs: list[MazeDatasetConfig] = serializable_field(
        serialization_fn=lambda configs: [config.serialize() for config in configs],
        loading_fn=lambda data: [
            MazeDatasetConfig.load(config) for config in data["maze_dataset_configs"]
        ],
    )

    def summary(self) -> dict:
        """return a summary of the config"""
        return [c.summary() for c in self.maze_dataset_configs]

    @property
    def n_mazes(self) -> int:
        return sum(config.n_mazes for config in self.maze_dataset_configs)

    @property
    def max_grid_n(self) -> int:
        return max(config.grid_n for config in self.maze_dataset_configs)

    @property
    def max_grid_shape(self) -> CoordTup:
        return (self.max_grid_n, self.max_grid_n)

    @property
    def max_grid_shape_np(self) -> Coord:
        return np.array(self.max_grid_shape, dtype=np.int32)

    def stable_hash_cfg(self) -> int:
        return stable_hash(json.dumps(self.serialize()))

    def to_fname(self) -> str:
        """convert config to a filename"""
        return sanitize_fname(
            f"collected-{self.name}-n{shorten_numerical_to_str(self.n_mazes)}-h{self.stable_hash_cfg()%10**5}"
        )


class MazeDatasetCollection(GPTDataset):
    """a collection of maze datasets"""

    def __init__(
        self,
        cfg: MazeDatasetCollectionConfig,
        maze_datasets: list[MazeDataset],
        generation_metadata_collected: dict | None = None,
    ) -> None:
        super().__init__()
        self.cfg: MazeDatasetCollectionConfig = cfg
        self.maze_datasets: list[MazeDataset] = list(maze_datasets)
        for c, ds in zip(self.cfg.maze_dataset_configs, self.maze_datasets):
            assert c.name == ds.cfg.name
            assert c == ds.cfg

        self.generation_metadata_collected: dict | None = generation_metadata_collected

    @property
    def dataset_lengths(self) -> list[int]:
        return [len(dataset) for dataset in self.maze_datasets]

    @property
    def dataset_cum_lengths(self) -> Int[np.ndarray, "indices"]:
        return np.array(list(itertools.accumulate(self.dataset_lengths)))

    @cached_property
    def mazes(self) -> list[LatticeMaze]:
        return list(
            itertools.chain.from_iterable(
                dataset.mazes for dataset in self.maze_datasets
            )
        )

    def __len__(self) -> int:
        return sum(len(dataset) for dataset in self.maze_datasets)

    def __getitem__(self, index: int):
        # find which dataset the index belongs to
        # we add 1, since np.searchsorted returns the
        # index of the last element that is strictly less than the target
        # while we want the index of the last element less than or equal to the target
        dataset_idx: int = np.searchsorted(self.dataset_cum_lengths, index + 1)
        index_adjusted: int = index
        if dataset_idx > 0:
            # if the index is 0, `dataset_idx - 1` will be -1.
            # We just want to use the base index
            index_adjusted -= self.dataset_cum_lengths[dataset_idx - 1]
        return self.maze_datasets[dataset_idx][index_adjusted]

    @classmethod
    def generate(
        cls, cfg: MazeDatasetCollectionConfig, **kwargs
    ) -> "MazeDatasetCollection":
        datasets = [
            MazeDataset.generate(config, **kwargs)
            for config in cfg.maze_dataset_configs
        ]
        return cls(cfg, datasets)

    @classmethod
    def download(
        cls, cfg: MazeDatasetCollectionConfig, **kwargs
    ) -> "MazeDatasetCollection":
        datasets = [
            MazeDataset.download(config, **kwargs)
            for config in cfg.maze_dataset_configs
        ]
        return cls(cfg, datasets)

    def serialize(self) -> JSONitem:
        return dict(
            __format__="MazeDatasetCollection",
            cfg=self.cfg.serialize(),
            maze_datasets=[dataset.serialize() for dataset in self.maze_datasets],
            generation_metadata_collected=json_serialize(
                self.generation_metadata_collected
            ),
        )

    @classmethod
    def load(cls, data: JSONitem) -> "MazeDatasetCollection":
        assert data["__format__"] == "MazeDatasetCollection"
        return cls(
            **{
                key: load_item_recursive(data[key], tuple())
                for key in ["cfg", "maze_datasets", "generation_metadata_collected"]
            }
        )

    # TODO: remove duplication with MazeDatasetConfig().as_tokens() somehow?
    def as_tokens(
        self,
        maze_tokenizer,  # TODO: MazeTokenizer
        limit: int | None = None,
        join_tokens_individual_maze: bool = False,
    ) -> list[list[str]] | list[str]:
        """return the dataset as tokens

        if join_tokens_individual_maze is True, then the tokens of each maze are
        joined with a space, and the result is a list of strings.
        i.e.:
        >>> dataset.as_tokens(join_tokens_individual_maze=False)
        [["a", "b", "c"], ["d", "e", "f"]]
        >>> dataset.as_tokens(join_tokens_individual_maze=True)
        ["a b c", "d e f"]
        """
        output: list[list[str]] = [
            maze.as_tokens(maze_tokenizer) for maze in self.mazes[:limit]
        ]
        if join_tokens_individual_maze:
            return [" ".join(tokens) for tokens in output]
        else:
            return output

    def update_self_config(self) -> None:
        # TODO: why cant we set this directly? its not frozen, and it seems to work in a regular MazeDataset
        self.cfg.__dict__["n_mazes"] = len(self)
        for dataset in self.maze_datasets:
            dataset.update_self_config()


MazeDatasetCollectionConfig._dataset_class = MazeDatasetCollection
register_loader_handler(
    LoaderHandler(
        check=lambda json_item, path=None, z=None: (
            isinstance(json_item, typing.Mapping)
            and "__format__" in json_item
            and json_item["__format__"].startswith("MazeDatasetCollection")
        ),
        load=lambda json_item, path=None, z=None: MazeDatasetCollection.load(json_item),
        uid="MazeDatasetCollection",
        source_pckg="maze_dataset.generation.maze_dataset_collection",
        desc="MazeDatasetCollection",
    )
)



from __future__ import annotations  # for type hinting self as return value

import warnings
from copy import deepcopy
from dataclasses import dataclass

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from jaxtyping import Bool, Float

from maze_dataset.constants import Coord, CoordArray, CoordList
from maze_dataset.maze import (
    Coord,
    CoordArray,
    LatticeMaze,
    SolvedMaze,
    TargetedLatticeMaze,
)

LARGE_NEGATIVE_NUMBER: float = -1e10


@dataclass(kw_only=True)
class PathFormat:
    """formatting options for path plot"""

    label: str | None = None
    fmt: str = "o"
    color: str | None = None
    cmap: str | None = None
    line_width: float | None = None
    quiver_kwargs: dict | None = None

    def combine(self, other: PathFormat) -> PathFormat:
        """combine with other PathFormat object, overwriting attributes with non-None values.

        returns a modified copy of self.
        """
        output: PathFormat = deepcopy(self)
        for key, value in other.__dict__.items():
            if key == "path":
                raise ValueError(
                    f"Cannot overwrite path attribute! {self = }, {other = }"
                )
            if value is not None:
                setattr(output, key, value)

        return output


# styled path
@dataclass
class StyledPath(PathFormat):
    path: CoordArray


DEFAULT_FORMATS: dict[str, PathFormat] = {
    "true": PathFormat(
        label="true path",
        fmt="--",
        color="red",
        line_width=2.5,
        quiver_kwargs=None,
    ),
    "predicted": PathFormat(
        label=None,
        fmt=":",
        color=None,
        line_width=2,
        quiver_kwargs={"width": 0.015},
    ),
}


def process_path_input(
    path: CoordList | CoordArray | StyledPath,
    _default_key: str,
    path_fmt: PathFormat | None = None,
    **kwargs,
) -> StyledPath:
    styled_path: StyledPath
    if isinstance(path, StyledPath):
        styled_path = path
    elif isinstance(path, np.ndarray):
        styled_path = StyledPath(path=path)
        # add default formatting
        styled_path = styled_path.combine(DEFAULT_FORMATS[_default_key])
    elif isinstance(path, list):
        styled_path = StyledPath(path=np.array(path))
        # add default formatting
        styled_path = styled_path.combine(DEFAULT_FORMATS[_default_key])
    else:
        raise TypeError(
            f"Expected CoordList, CoordArray or StyledPath, got {type(path)}: {path}"
        )

    # add formatting from path_fmt
    if path_fmt is not None:
        styled_path = styled_path.combine(path_fmt)

    # add formatting from kwargs
    for key, value in kwargs.items():
        setattr(styled_path, key, value)

    return styled_path


class MazePlot:
    """Class for displaying mazes and paths"""

    DEFAULT_PREDICTED_PATH_COLORS: list[str] = [
        "tab:orange",
        "tab:olive",
        "sienna",
        "mediumseagreen",
        "tab:purple",
        "slategrey",
    ]

    def __init__(self, maze: LatticeMaze, unit_length: int = 14) -> None:
        """
        UNIT_LENGTH: Set ratio between node size and wall thickness in image.
        Wall thickness is fixed to 1px
        A "unit" consists of a single node and the right and lower connection/wall.
        Example: ul = 14 yields 13:1 ratio between node size and wall thickness
        """
        self.unit_length: int = unit_length
        self.maze: LatticeMaze = maze
        self.true_path: StyledPath | None = None
        self.predicted_paths: list[StyledPath] = []
        self.node_values: Float[np.ndarray, "grid_n grid_n"] = None
        self.custom_node_value_flag: bool = False
        self.node_color_map: str = "Blues"
        self.target_token_coord: Coord = None
        self.preceding_tokens_coords: CoordArray = None
        self.colormap_center: float | None = None
        self.cbar_ax = None
        self.marked_coords: list[tuple[Coord, dict]] = list()

        self.marker_kwargs_current: dict = dict(
            marker="s",
            color="green",
            ms=12,
        )
        self.marker_kwargs_next: dict = dict(
            marker="P",
            color="green",
            ms=12,
        )

        if isinstance(maze, SolvedMaze):
            self.add_true_path(maze.solution)
        else:
            if isinstance(maze, TargetedLatticeMaze):
                self.add_true_path(SolvedMaze.from_targeted_lattice_maze(maze).solution)

    @property
    def solved_maze(self) -> SolvedMaze:
        if self.true_path is None:
            raise ValueError(
                "Cannot return SolvedMaze object without true path. Add true path with add_true_path method."
            )
        return SolvedMaze.from_lattice_maze(
            lattice_maze=self.maze,
            solution=self.true_path.path,
        )

    def add_true_path(
        self,
        path: CoordList | CoordArray | StyledPath,
        path_fmt: PathFormat | None = None,
        **kwargs,
    ) -> MazePlot:
        self.true_path = process_path_input(
            path=path,
            _default_key="true",
            path_fmt=path_fmt,
            **kwargs,
        )

        return self

    def add_predicted_path(
        self,
        path: CoordList | CoordArray | StyledPath,
        path_fmt: PathFormat | None = None,
        **kwargs,
    ) -> MazePlot:
        """
        Recieve predicted path and formatting preferences from input and save in predicted_path list.
        Default formatting depends on nuber of paths already saved in predicted path list.
        """
        styled_path: StyledPath = process_path_input(
            path=path,
            _default_key="predicted",
            path_fmt=path_fmt,
            **kwargs,
        )

        # set default label and color if not specified
        if styled_path.label is None:
            styled_path.label = f"predicted path {len(self.predicted_paths) + 1}"

        if styled_path.color is None:
            color_num: int = len(self.predicted_paths) % len(
                self.DEFAULT_PREDICTED_PATH_COLORS
            )
            styled_path.color = self.DEFAULT_PREDICTED_PATH_COLORS[color_num]

        self.predicted_paths.append(styled_path)
        return self

    def add_multiple_paths(self, path_list: list[CoordList | CoordArray | StyledPath]):
        """
        Function for adding multiple paths to MazePlot at once. This can be done in two ways:
        1. Passing a list of
        """
        for path in path_list:
            self.add_predicted_path(path)
        return self

    def add_node_values(
        self,
        node_values: Float[np.ndarray, "grid_n grid_n"],
        color_map: str = "Blues",
        target_token_coord: Coord | None = None,
        preceeding_tokens_coords: CoordArray = None,
        colormap_center: float | None = None,
        colormap_max: float | None = None,
        hide_colorbar: bool = False,
    ) -> MazePlot:
        assert (
            node_values.shape == self.maze.grid_shape
        ), "Please pass node values of the same sape as LatticeMaze.grid_shape"
        # assert np.min(node_values) >= 0, "Please pass non-negative node values only."

        self.node_values = node_values
        # Set flag for choosing cmap while plotting maze
        self.custom_node_value_flag = True
        # Retrieve Max node value for plotting, +1e-10 to avoid division by zero
        self.node_color_map = color_map
        self.colormap_center = colormap_center
        self.colormap_max = colormap_max
        self.hide_colorbar = hide_colorbar

        if target_token_coord is not None:
            self.marked_coords.append((target_token_coord, self.marker_kwargs_next))
        if preceeding_tokens_coords is not None:
            for coord in preceeding_tokens_coords:
                self.marked_coords.append((coord, self.marker_kwargs_current))
        return self

    def plot(
        self,
        dpi: int = 100,
        title: str = "",
        fig_ax: tuple | None = None,
        plain: bool = False,
    ) -> MazePlot:
        """Plot the maze and paths."""

        # set up figure
        if fig_ax is None:
            self.fig = plt.figure(dpi=dpi)
            self.ax = self.fig.add_subplot(1, 1, 1)
        else:
            self.fig, self.ax = fig_ax

        # plot maze
        self._plot_maze()

        # Plot labels
        if not plain:
            tick_arr = np.arange(self.maze.grid_shape[0])
            self.ax.set_xticks(self.unit_length * (tick_arr + 0.5), tick_arr)
            self.ax.set_yticks(self.unit_length * (tick_arr + 0.5), tick_arr)
            self.ax.set_xlabel("col")
            self.ax.set_ylabel("row")
            self.ax.set_title(title)

        # plot paths
        if self.true_path is not None:
            self._plot_path(self.true_path)
        for path in self.predicted_paths:
            self._plot_path(path)

        # plot markers
        for coord, kwargs in self.marked_coords:
            self._place_marked_coords([coord], **kwargs)

    def _rowcol_to_coord(self, point: Coord) -> np.ndarray:
        """Transform Point from MazeTransformer (row, column) notation to matplotlib default (x, y) notation where x is the horizontal axis."""
        point = np.array([point[1], point[0]])
        return self.unit_length * (point + 0.5)

    def mark_coords(self, coords: CoordArray | list[Coord], **kwargs) -> MazePlot:
        kwargs = {
            **dict(marker="+", color="blue"),
            **kwargs,
        }
        for coord in coords:
            self.marked_coords.append((coord, kwargs))

    def _place_marked_coords(
        self, coords: CoordArray | list[Coord], **kwargs
    ) -> MazePlot:
        coords_tp = np.array([self._rowcol_to_coord(coord) for coord in coords])
        self.ax.plot(coords_tp[:, 0], coords_tp[:, 1], **kwargs)

    def _plot_maze(self) -> None:
        """
        Define Colormap and plot maze.
        Colormap: x is -inf: black
                  else: use colormap
        """
        img = self._lattice_maze_to_img()

        # if no node_values have been passed (no colormap)
        if self.custom_node_value_flag is False:
            self.ax.imshow(img, cmap="gray", vmin=-1, vmax=1)

        else:
            assert self.node_values is not None, "Please pass node values."
            assert not np.isnan(
                self.node_values
            ).any(), "Please pass node values, they cannot be nan."

            vals_min: float = np.nanmin(self.node_values)
            vals_max: float = np.nanmax(self.node_values)
            # if both are negative or both are positive, set max/min to 0
            if vals_max < 0.0:
                vals_max = 0.0
            elif vals_min > 0.0:
                vals_min = 0.0

            # adjust vals_max, in case you need consistent colorbar across multiple plots
            vals_max: float = self.colormap_max or vals_max

            # create colormap
            cmap = mpl.colormaps[self.node_color_map]
            # TODO: this is a hack, we make the walls black (while still allowing negative values) by setting the nan color to black
            cmap.set_bad(color="black")

            if self.colormap_center is not None:
                if not (vals_min < self.colormap_center < vals_max):
                    if vals_min == self.colormap_center:
                        vals_min -= 1e-10
                    elif vals_max == self.colormap_center:
                        vals_max += 1e-10
                    else:
                        raise ValueError(
                            f"Please pass colormap_center value between {vals_min} and {vals_max}"
                        )

                norm = mpl.colors.TwoSlopeNorm(
                    vmin=vals_min,
                    vcenter=self.colormap_center,
                    vmax=vals_max,
                )
                _plotted = self.ax.imshow(img, cmap=cmap, norm=norm)
            else:
                _plotted = self.ax.imshow(img, cmap=cmap, vmin=vals_min, vmax=vals_max)

            # Add colorbar based on the condition of self.hide_colorbar
            if not self.hide_colorbar:
                ticks = np.linspace(vals_min, vals_max, 5)

                if (vals_min < 0.0 < vals_max) and (0.0 not in ticks):
                    ticks = np.insert(ticks, np.searchsorted(ticks, 0.0), 0.0)

                if (
                    self.colormap_center is not None
                    and self.colormap_center not in ticks
                    and vals_min < self.colormap_center < vals_max
                ):
                    ticks = np.insert(
                        ticks,
                        np.searchsorted(ticks, self.colormap_center),
                        self.colormap_center,
                    )

                cbar = plt.colorbar(
                    _plotted,
                    ticks=ticks,
                    ax=self.ax,
                    cax=self.cbar_ax,
                )
                self.cbar_ax = cbar.ax

        # make the boundaries of the image thicker (walls look weird without this)
        for axis in ["top", "bottom", "left", "right"]:
            self.ax.spines[axis].set_linewidth(2)

    def _lattice_maze_to_img(
        self,
        connection_val_scale: float = 0.93,
    ) -> Bool[np.ndarray, "row col"]:
        """
        Build an image to visualise the maze.
        Each "unit" consists of a node and the right and lower adjacent wall/connection. Its area is ul * ul.
        - Nodes have area: (ul-1) * (ul-1) and value 1 by default
            - take node_value if passed via .add_node_values()
        - Walls have area: 1 * (ul-1) and value -1
        - Connections have area: 1 * (ul-1); color and value 0.93 by default
            - take node_value if passed via .add_node_values()

        Axes definition:
        (0,0)     col
        ----|----------->
            |
        row |
            |
            v

        Returns a matrix of side length (ul) * n + 1 where n is the number of nodes.
        """

        # TODO: this is a hack, but if you add 1 always then non-node valued plots have their walls dissapear. if you dont add 1, you get ugly colors between nodes when they are colored
        node_bdry_hack: int
        connection_list_processed: Float[np.ndarray, "dim row col"]
        # Set node and connection values
        if self.node_values is None:
            scaled_node_values = np.ones(self.maze.grid_shape)
            connection_values = scaled_node_values * connection_val_scale
            node_bdry_hack = 0
            # TODO: hack
            # invert connection list
            connection_list_processed = np.logical_not(self.maze.connection_list)
        else:
            # TODO: hack
            scaled_node_values = self.node_values
            # connection_values = scaled_node_values
            connection_values = np.full_like(scaled_node_values, np.nan)
            node_bdry_hack = 1
            connection_list_processed = self.maze.connection_list

        # Create background image (all pixels set to -1, walls everywhere)
        img: Float[np.ndarray, "row col"] = -np.ones(
            (
                self.maze.grid_shape[0] * self.unit_length + 1,
                self.maze.grid_shape[1] * self.unit_length + 1,
            ),
            dtype=float,
        )

        # Draw nodes and connections by iterating through lattice
        for row in range(self.maze.grid_shape[0]):
            for col in range(self.maze.grid_shape[1]):
                # Draw node
                img[
                    row * self.unit_length
                    + 1 : (row + 1) * self.unit_length
                    + node_bdry_hack,
                    col * self.unit_length
                    + 1 : (col + 1) * self.unit_length
                    + node_bdry_hack,
                ] = scaled_node_values[row, col]

                # Down connection
                if not connection_list_processed[0, row, col]:
                    img[
                        (row + 1) * self.unit_length,
                        col * self.unit_length + 1 : (col + 1) * self.unit_length,
                    ] = connection_values[row, col]

                # Right connection
                if not connection_list_processed[1, row, col]:
                    img[
                        row * self.unit_length + 1 : (row + 1) * self.unit_length,
                        (col + 1) * self.unit_length,
                    ] = connection_values[row, col]

        return img

    def _plot_path(self, path_format: PathFormat) -> None:
        if len(path_format.path) == 0:
            warnings.warn(f"Empty path, skipping plotting\n{path_format = }")
            return
        p_transformed = np.array(
            [self._rowcol_to_coord(coord) for coord in path_format.path]
        )
        if path_format.quiver_kwargs is not None:
            try:
                x: np.ndarray = p_transformed[:, 0]
                y: np.ndarray = p_transformed[:, 1]
            except Exception as e:
                raise ValueError(
                    f"Error in plotting quiver path:\n{path_format = }\n{p_transformed = }\n{e}"
                ) from e

            # Generate colors from the colormap
            if path_format.cmap is not None:
                n = len(x) - 1  # Number of arrows
                cmap = plt.get_cmap(path_format.cmap)
                colors = [cmap(i / n) for i in range(n)]
            else:
                colors = path_format.color

            self.ax.quiver(
                x[:-1],
                y[:-1],
                x[1:] - x[:-1],
                y[1:] - y[:-1],
                scale_units="xy",
                angles="xy",
                scale=1,
                color=colors,
                **path_format.quiver_kwargs,
            )
        else:
            x: np.ndarray = p_transformed[:, 0]
            y: np.ndarray = p_transformed[:, 1]
            self.ax.plot(
                x,
                y,
                path_format.fmt,
                lw=path_format.line_width,
                color=path_format.color,
                label=path_format.label,
            )
        # mark endpoints
        self.ax.plot(
            [p_transformed[0][0]],
            [p_transformed[0][1]],
            "o",
            color=path_format.color,
            ms=10,
        )
        self.ax.plot(
            [p_transformed[-1][0]],
            [p_transformed[-1][1]],
            "x",
            color=path_format.color,
            ms=10,
        )

    def to_ascii(
        self,
        show_endpoints: bool = True,
        show_solution: bool = True,
    ) -> str:
        if self.true_path:
            return self.solved_maze.as_ascii(
                show_endpoints=show_endpoints, show_solution=show_solution
            )
        else:
            return self.maze.as_ascii(show_endpoints=show_endpoints)



from maze_dataset.constants import (
    SPECIAL_TOKENS,
    Coord,
    CoordArray,
    CoordList,
    CoordTup,
)
from maze_dataset.dataset.collected_dataset import (
    MazeDatasetCollection,
    MazeDatasetCollectionConfig,
)
from maze_dataset.dataset.maze_dataset import (
    MazeDataset,
    MazeDatasetConfig,
    set_serialize_minimal_threshold,
)
from maze_dataset.generation.generators import LatticeMazeGenerators
from maze_dataset.maze.lattice_maze import LatticeMaze, SolvedMaze

__all__ = [
    "Coord",
    "CoordTup",
    "CoordList",
    "CoordArray",
    "SPECIAL_TOKENS",
    "LatticeMaze",
    "SolvedMaze",
    "MazeDataset",
    "MazeDatasetConfig",
    "set_serialize_minimal_threshold",
    "MazeDatasetCollection",
    "MazeDatasetCollectionConfig",
    "LatticeMazeGenerators",
]



import matplotlib.pyplot as plt  # type: ignore[import]

from maze_dataset.dataset.maze_dataset import MazeDataset


def plot_dataset_mazes(
    ds: MazeDataset,
    count: int | None = None,
    figsize_mult: tuple[float, float] = (1.0, 2.0),
    title: bool | str = True,
) -> tuple:
    count = count or len(ds)
    if count == 0:
        print(f"No mazes to plot for dataset")
        return
    fig, axes = plt.subplots(
        1, count, figsize=(count * figsize_mult[0], figsize_mult[1])
    )
    if count == 1:
        axes = [axes]
    for i in range(count):
        axes[i].imshow(ds[i].as_pixels())
        # remove ticks
        axes[i].set_xticks([])
        axes[i].set_yticks([])

    # set title
    if title:
        if isinstance(title, str):
            fig.suptitle(title)
        else:
            kwargs: dict = {
                "grid_n": ds.cfg.grid_n,
                # "n_mazes": ds.cfg.n_mazes,
                **ds.cfg.maze_ctor_kwargs,
            }
            fig.suptitle(
                f"{ds.cfg.to_fname()}\n{ds.cfg.maze_ctor.__name__}({', '.join(f'{k}={v}' for k, v in kwargs.items())})"
            )

    # tight layout
    fig.tight_layout()
    # remove whitespace between title and subplots
    fig.subplots_adjust(top=1.0)

    return fig, axes


def print_dataset_mazes(ds: MazeDataset, count: int | None = None):
    count = count or len(ds)
    if count == 0:
        print(f"No mazes to print for dataset")
        return
    for i in range(count):
        print(ds[i].as_ascii(), "\n\n-----\n")



from maze_dataset.dataset.maze_dataset import MazeDatasetConfig
from maze_dataset.generation.generators import LatticeMazeGenerators

MAZE_DATASET_CONFIGS: dict[str, MazeDatasetConfig] = {
    cfg.to_fname(): cfg
    for cfg in [
        MazeDatasetConfig(
            name="test",
            grid_n=3,
            n_mazes=5,
            maze_ctor=LatticeMazeGenerators.gen_dfs,
        ),
        MazeDatasetConfig(
            name="demo_small",
            grid_n=3,
            n_mazes=100,
            maze_ctor=LatticeMazeGenerators.gen_dfs,
        ),
        MazeDatasetConfig(
            name="demo",
            grid_n=6,
            n_mazes=10000,
            maze_ctor=LatticeMazeGenerators.gen_dfs,
        ),
    ]
}



"""DEFAULT_GENERATORS is a list of generator name, generator kwargs pairs"""

DEFAULT_GENERATORS: list[tuple[str, dict]] = [
    ("gen_dfs", dict()),
    ("gen_dfs", dict(do_forks=False)),
    ("gen_dfs", dict(accessible_cells=20)),
    ("gen_dfs", dict(max_tree_depth=0.5)),
    ("gen_wilson", dict()),
    # ("gen_percolation", dict(p=0.1)),
    (
        "gen_percolation",
        dict(p=1.0),
    ),  # anything less than this and tests will stochastically fail
    ("gen_dfs_percolation", dict(p=0.1)),
    ("gen_dfs_percolation", dict(p=0.4)),
    # ("gen_prim", dict()),
    # ("gen_prim", dict(do_forks=False)),
    # ("gen_prim", dict(accessible_cells=0.5)),
    # ("gen_prim", dict(max_tree_depth=0.5)),
    # ("gen_prim", dict(accessible_cells=0.5, max_tree_depth=0.5)),
]



from maze_dataset.plotting.plot_dataset import plot_dataset_mazes, print_dataset_mazes
from maze_dataset.plotting.plot_maze import DEFAULT_FORMATS, MazePlot, PathFormat

__all__ = [
    "plot_dataset_mazes",
    "print_dataset_mazes",
    "DEFAULT_FORMATS",
    "MazePlot",
    "PathFormat",
]



import warnings
from dataclasses import dataclass

import numpy as np
from jaxtyping import Int8

Coord = Int8[np.ndarray, "x y"]
CoordTup = tuple[int, int]
CoordArray = Int8[np.ndarray, "coord x y"]
CoordList = list[CoordTup]


class SpecialTokensError(Exception):
    pass


_SPECIAL_TOKENS_ABBREVIATIONS: dict[str, str] = {
    "<ADJLIST_START>": "<A_S>",
    "<ADJLIST_END>": "<A_E>",
    "<TARGET_START>": "<T_S>",
    "<TARGET_END>": "<T_E>",
    "<ORIGIN_START>": "<O_S>",
    "<ORIGIN_END>": "<O_E>",
    "<PATH_START>": "<P_S>",
    "<PATH_END>": "<P_E>",
    "<-->": "<-->",
    ";": ";",
    "<PADDING>": "<PAD>",
}


@dataclass(frozen=True)
class _SPECIAL_TOKENS_BASE:
    ADJLIST_START: str = "<ADJLIST_START>"
    ADJLIST_END: str = "<ADJLIST_END>"
    TARGET_START: str = "<TARGET_START>"
    TARGET_END: str = "<TARGET_END>"
    ORIGIN_START: str = "<ORIGIN_START>"
    ORIGIN_END: str = "<ORIGIN_END>"
    PATH_START: str = "<PATH_START>"
    PATH_END: str = "<PATH_END>"
    CONNECTOR: str = "<-->"
    ADJACENCY_ENDLINE: str = ";"
    PADDING: str = "<PADDING>"

    def __getitem__(self, key: str) -> str:
        key_upper: str = key.upper()

        if not isinstance(key, str):
            raise TypeError(f"key must be str, not {type(key)}")

        # error checking for old lowercase format
        if not key == key_upper:
            warnings.warn(
                f"Accessing special token '{key}' without uppercase. this is deprecated and will be removed in the future.",
                DeprecationWarning,
            )
            key = key_upper

        # `ADJLIST` used to be `adj_list`, changed to match actual token content
        if not key_upper in self.keys():
            key_upper_modified: str = key_upper.replace("ADJ_LIST", "ADJLIST")
            if key_upper_modified in self.keys():
                warnings.warn(
                    f"Accessing '{key}' in old format, should use {key_upper_modified}. this is deprecated and will be removed in the future.",
                    DeprecationWarning,
                )
                return getattr(self, key_upper_modified)
            else:
                raise KeyError(f"invalid special token '{key}'")

        # normal return
        return getattr(self, key.upper())

    def get_abbrev(self, key: str) -> str:
        return _SPECIAL_TOKENS_ABBREVIATIONS[self[key]]

    def __iter__(self):
        return iter(self.__dict__.keys())

    def __len__(self):
        return len(self.__dict__.keys())

    def __contains__(self, key: str) -> bool:
        return key in self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def items(self):
        return self.__dict__.items()

    def keys(self):
        return self.__dict__.keys()


SPECIAL_TOKENS: _SPECIAL_TOKENS_BASE = _SPECIAL_TOKENS_BASE()


DIRECTIONS_MAP: Int8[np.ndarray, "direction axes"] = np.array(
    [
        [0, 1],  # down
        [0, -1],  # up
        [1, 1],  # right
        [1, -1],  # left
    ]
)


NEIGHBORS_MASK: Int8[np.ndarray, "coord point"] = np.array(
    [
        [0, 1],  # down
        [0, -1],  # up
        [1, 0],  # right
        [-1, 0],  # left
    ]
)



import random
import warnings
from typing import Any, Callable

import numpy as np
from jaxtyping import Bool

from maze_dataset.constants import CoordArray
from maze_dataset.maze import ConnectionList, Coord, LatticeMaze, SolvedMaze
from maze_dataset.maze.lattice_maze import NEIGHBORS_MASK, _fill_edges_with_walls


def _random_start_coord(grid_shape: Coord, start_coord: Coord | None) -> Coord:
    if start_coord is None:
        start_coord: Coord = np.random.randint(
            0,  # lower bound
            np.maximum(grid_shape - 1, 1),  # upper bound (at least 1)
            size=len(grid_shape),  # dimensionality
        )
    else:
        start_coord = np.array(start_coord)

    return start_coord


def get_neighbors_in_bounds(
    coord: Coord,
    grid_shape: Coord,
) -> CoordArray:
    # get all neighbors
    neighbors: CoordArray = coord + NEIGHBORS_MASK

    # filter neighbors by being within grid bounds
    neighbors_in_bounds: CoordArray = neighbors[
        (neighbors >= 0).all(axis=1) & (neighbors < grid_shape).all(axis=1)
    ]

    return neighbors_in_bounds


class LatticeMazeGenerators:
    """namespace for lattice maze generation algorithms"""

    @staticmethod
    def gen_dfs(
        grid_shape: Coord,
        lattice_dim: int = 2,
        accessible_cells: int | float | None = None,
        max_tree_depth: int | float | None = None,
        do_forks: bool = True,
        randomized_stack: bool = False,
        start_coord: Coord | None = None,
    ) -> LatticeMaze:
        """generate a lattice maze using depth first search, iterative

        # Arguments
        - `grid_shape: Coord`: the shape of the grid
        - `lattice_dim: int`: the dimension of the lattice
          (default: `2`)
        - `accessible_cells: int | float |None`: the number of accessible cells in the maze. If `None`, defaults to the total number of cells in the grid. if a float, asserts it is <= 1 and treats it as a proportion of **total cells**
            (default: `None`)
        - `max_tree_depth: int | float | None`: the maximum depth of the tree. If `None`, defaults to `2 * accessible_cells`. if a float, asserts it is <= 1 and treats it as a proportion of the **sum of the grid shape**
            (default: `None`)
        - `do_forks: bool`: whether to allow forks in the maze. If `False`, the maze will be have no forks and will be a simple hallway.
        - `start_coord: Coord | None`: the starting coordinate of the generation algorithm. If `None`, defaults to a random coordinate.

        # algorithm
        1. Choose the initial cell, mark it as visited and push it to the stack
        2. While the stack is not empty
                1. Pop a cell from the stack and make it a current cell
                2. If the current cell has any neighbours which have not been visited
                        1. Push the current cell to the stack
                        2. Choose one of the unvisited neighbours
                        3. Remove the wall between the current cell and the chosen cell
                        4. Mark the chosen cell as visited and push it to the stack
        """

        # Default values if no constraints have been passed
        grid_shape: Coord = np.array(grid_shape)
        n_total_cells: int = int(np.prod(grid_shape))

        n_accessible_cells: int
        if accessible_cells is None:
            n_accessible_cells = n_total_cells
        elif isinstance(accessible_cells, float):
            assert (
                accessible_cells <= 1
            ), f"accessible_cells must be an int (count) or a float in the range [0, 1] (proportion), got {accessible_cells}"

            n_accessible_cells = int(accessible_cells * n_total_cells)
        else:
            assert isinstance(accessible_cells, int)
            n_accessible_cells = accessible_cells

        if max_tree_depth is None:
            max_tree_depth = (
                2 * n_total_cells
            )  # We define max tree depth counting from the start coord in two directions. Therefore we divide by two in the if clause for neighboring sites later and multiply by two here.
        elif isinstance(max_tree_depth, float):
            assert (
                max_tree_depth <= 1
            ), f"max_tree_depth must be an int (count) or a float in the range [0, 1] (proportion), got {max_tree_depth}"

            max_tree_depth = int(max_tree_depth * np.sum(grid_shape))

        # choose a random start coord
        start_coord = _random_start_coord(grid_shape, start_coord)

        # initialize the maze with no connections
        connection_list: ConnectionList = np.zeros(
            (lattice_dim, grid_shape[0], grid_shape[1]), dtype=np.bool_
        )

        # initialize the stack with the target coord
        visited_cells: set[tuple[int, int]] = set()
        visited_cells.add(tuple(start_coord))  # this wasnt a bug after all lol
        stack: list[Coord] = [start_coord]

        # initialize tree_depth_counter
        current_tree_depth: int = 1

        # loop until the stack is empty or n_connected_cells is reached
        while stack and (len(visited_cells) < n_accessible_cells):
            # get the current coord from the stack
            current_coord: Coord
            if randomized_stack:
                current_coord = stack.pop(random.randint(0, len(stack) - 1))
            else:
                current_coord = stack.pop()

            # filter neighbors by being within grid bounds and being unvisited
            unvisited_neighbors_deltas: list[tuple[Coord, Coord]] = [
                (neighbor, delta)
                for neighbor, delta in zip(
                    current_coord + NEIGHBORS_MASK, NEIGHBORS_MASK
                )
                if (
                    (tuple(neighbor) not in visited_cells)
                    and (0 <= neighbor[0] < grid_shape[0])
                    and (0 <= neighbor[1] < grid_shape[1])
                )
            ]

            # don't continue if max_tree_depth/2 is already reached (divide by 2 because we can branch to multiple directions)
            if unvisited_neighbors_deltas and (
                current_tree_depth <= max_tree_depth / 2
            ):
                # if we want a maze without forks, simply don't add the current coord back to the stack
                if do_forks and (len(unvisited_neighbors_deltas) > 1):
                    stack.append(current_coord)

                # choose one of the unvisited neighbors
                chosen_neighbor, delta = random.choice(unvisited_neighbors_deltas)

                # add connection
                dim: int = np.argmax(np.abs(delta))
                # if positive, down/right from current coord
                # if negative, up/left from current coord (down/right from neighbor)
                clist_node: Coord = (
                    current_coord if (delta.sum() > 0) else chosen_neighbor
                )
                connection_list[dim, clist_node[0], clist_node[1]] = True

                # add to visited cells and stack
                visited_cells.add(tuple(chosen_neighbor))
                stack.append(chosen_neighbor)

                # Update current tree depth
                current_tree_depth += 1
            else:
                current_tree_depth -= 1

        output = LatticeMaze(
            connection_list=connection_list,
            generation_meta=dict(
                func_name="gen_dfs",
                grid_shape=grid_shape,
                start_coord=start_coord,
                n_accessible_cells=int(n_accessible_cells),
                max_tree_depth=int(max_tree_depth),
                # oh my god this took so long to track down. its almost 5am and I've spent like 2 hours on this bug
                # it was checking that len(visited_cells) == n_accessible_cells, but this means that the maze is
                # treated as fully connected even when it is most certainly not, causing solving the maze to break
                fully_connected=bool(len(visited_cells) == n_total_cells),
                visited_cells={tuple(int(x) for x in coord) for coord in visited_cells},
            ),
        )

        return output

    @staticmethod
    def gen_prim(
        grid_shape: Coord,
        lattice_dim: int = 2,
        accessible_cells: int | float | None = None,
        max_tree_depth: int | float | None = None,
        do_forks: bool = True,
        start_coord: Coord | None = None,
    ) -> LatticeMaze:
        warnings.warn(
            "gen_prim does not correctly implement prim's algorithm, see issue: https://github.com/understanding-search/maze-dataset/issues/12",
        )
        return LatticeMazeGenerators.gen_dfs(
            grid_shape=grid_shape,
            lattice_dim=lattice_dim,
            accessible_cells=accessible_cells,
            max_tree_depth=max_tree_depth,
            do_forks=do_forks,
            start_coord=start_coord,
            randomized_stack=True,
        )

    @staticmethod
    def gen_wilson(
        grid_shape: Coord,
    ) -> LatticeMaze:
        """Generate a lattice maze using Wilson's algorithm.

        # Algorithm
        Wilson's algorithm generates an unbiased (random) maze
        sampled from the uniform distribution over all mazes, using loop-erased random walks. The generated maze is
        acyclic and all cells are part of a unique connected space.
        https://en.wikipedia.org/wiki/Maze_generation_algorithm#Wilson's_algorithm
        """

        # Initialize grid and visited cells
        connection_list: ConnectionList = np.zeros((2, *grid_shape), dtype=np.bool_)
        visited: Bool[np.ndarray, "x y"] = np.zeros(grid_shape, dtype=np.bool_)

        # Choose a random cell and mark it as visited
        start_coord: Coord = _random_start_coord(grid_shape, None)
        visited[start_coord[0], start_coord[1]] = True
        del start_coord

        while not visited.all():
            # Perform loop-erased random walk from another random cell

            # Choose walk_start only from unvisited cells
            unvisited_coords: CoordArray = np.column_stack(np.where(~visited))
            walk_start: Coord = unvisited_coords[
                np.random.choice(unvisited_coords.shape[0])
            ]

            # Perform the random walk
            path: list[Coord] = [walk_start]
            current: Coord = walk_start

            # exit the loop once the current path hits a visited cell
            while not visited[current[0], current[1]]:
                # find a valid neighbor (one always exists on a lattice)
                neighbors: CoordArray = get_neighbors_in_bounds(current, grid_shape)
                next_cell: Coord = neighbors[np.random.choice(neighbors.shape[0])]

                # Check for loop
                loop_exit: int | None = None
                for i, p in enumerate(path):
                    if np.array_equal(next_cell, p):
                        loop_exit = i
                        break

                # erase the loop, or continue the walk
                if loop_exit is not None:
                    # this removes everything after and including the loop start
                    path = path[: loop_exit + 1]
                    # reset current cell to end of path
                    current = path[-1]
                else:
                    path.append(next_cell)
                    current = next_cell

            # Add the path to the maze
            for i in range(len(path) - 1):
                c_1: Coord = path[i]
                c_2: Coord = path[i + 1]

                # find the dimension of the connection
                delta: Coord = c_2 - c_1
                dim: int = np.argmax(np.abs(delta))

                # if positive, down/right from current coord
                # if negative, up/left from current coord (down/right from neighbor)
                clist_node: Coord = c_1 if (delta.sum() > 0) else c_2
                connection_list[dim, clist_node[0], clist_node[1]] = True
                visited[c_1[0], c_1[1]] = True
                # we dont add c_2 because the last c_2 will have already been visited

        return LatticeMaze(
            connection_list=connection_list,
            generation_meta=dict(
                func_name="gen_wilson",
                grid_shape=grid_shape,
                fully_connected=True,
            ),
        )

    @staticmethod
    def gen_percolation(
        grid_shape: Coord,
        p: float = 0.4,
        lattice_dim: int = 2,
        start_coord: Coord | None = None,
    ) -> LatticeMaze:
        """generate a lattice maze using simple percolation

        note that p in the range (0.4, 0.7) gives the most interesting mazes

        # Arguments
        - `grid_shape: Coord`: the shape of the grid
        - `lattice_dim: int`: the dimension of the lattice (default: `2`)
        - `p: float`: the probability of a cell being accessible (default: `0.5`)
        - `start_coord: Coord | None`: the starting coordinate for the connected component (default: `None` will give a random start)
        """
        assert p >= 0 and p <= 1, f"p must be between 0 and 1, got {p}"
        grid_shape: Coord = np.array(grid_shape)

        start_coord = _random_start_coord(grid_shape, start_coord)

        connection_list: ConnectionList = np.random.rand(lattice_dim, *grid_shape) < p

        connection_list = _fill_edges_with_walls(connection_list)

        output: LatticeMaze = LatticeMaze(
            connection_list=connection_list,
            generation_meta=dict(
                func_name="gen_percolation",
                grid_shape=grid_shape,
                percolation_p=p,
                start_coord=start_coord,
            ),
        )

        output.generation_meta["visited_cells"] = output.gen_connected_component_from(
            start_coord
        )

        return output

    @staticmethod
    def gen_dfs_percolation(
        grid_shape: Coord,
        p: float = 0.4,
        lattice_dim: int = 2,
        accessible_cells: int | None = None,
        max_tree_depth: int | None = None,
        start_coord: Coord | None = None,
    ) -> LatticeMaze:
        """dfs and then percolation (adds cycles)"""
        grid_shape: Coord = np.array(grid_shape)
        start_coord = _random_start_coord(grid_shape, start_coord)

        # generate initial maze via dfs
        maze: LatticeMaze = LatticeMazeGenerators.gen_dfs(
            grid_shape=grid_shape,
            lattice_dim=lattice_dim,
            accessible_cells=accessible_cells,
            max_tree_depth=max_tree_depth,
            start_coord=start_coord,
        )

        # percolate
        connection_list_perc: np.ndarray = (
            np.random.rand(*maze.connection_list.shape) < p
        )
        connection_list_perc = _fill_edges_with_walls(connection_list_perc)

        maze.__dict__["connection_list"] = np.logical_or(
            maze.connection_list, connection_list_perc
        )

        maze.generation_meta["func_name"] = "gen_dfs_percolation"
        maze.generation_meta["percolation_p"] = p
        maze.generation_meta["visited_cells"] = maze.gen_connected_component_from(
            start_coord
        )

        return maze


# cant automatically populate this because it messes with pickling :(
GENERATORS_MAP: dict[str, Callable[[Coord, Any], "LatticeMaze"]] = {
    "gen_dfs": LatticeMazeGenerators.gen_dfs,
    "gen_wilson": LatticeMazeGenerators.gen_wilson,
    "gen_percolation": LatticeMazeGenerators.gen_percolation,
    "gen_dfs_percolation": LatticeMazeGenerators.gen_dfs_percolation,
    "gen_prim": LatticeMazeGenerators.gen_prim,
}


def get_maze_with_solution(
    gen_name: str,
    grid_shape: Coord,
    maze_ctor_kwargs: dict | None = None,
) -> SolvedMaze:
    if maze_ctor_kwargs is None:
        maze_ctor_kwargs = dict()
    maze: LatticeMaze = GENERATORS_MAP[gen_name](grid_shape, **maze_ctor_kwargs)
    solution: CoordArray = np.array(maze.generate_random_path())
    return SolvedMaze.from_lattice_maze(lattice_maze=maze, solution=solution)



import copy
import functools
import json
import multiprocessing
import typing
import warnings
from collections import Counter, defaultdict
from typing import Callable

import numpy as np
import tqdm
from jaxtyping import Int
from muutils.json_serialize import (
    JSONitem,
    json_serialize,
    serializable_dataclass,
    serializable_field,
)
from muutils.json_serialize.util import safe_getsource, string_as_lines
from muutils.misc import sanitize_fname, shorten_numerical_to_str, stable_hash
from zanj.loading import LoaderHandler, load_item_recursive, register_loader_handler

from maze_dataset.constants import Coord, CoordTup
from maze_dataset.dataset.dataset import (
    DatasetFilterProtocol,
    GPTDataset,
    GPTDatasetConfig,
    register_dataset_filter,
    register_filter_namespace_for_dataset,
)
from maze_dataset.generation.generators import GENERATORS_MAP
from maze_dataset.maze import LatticeMaze, SolvedMaze

# If `n_mazes>=SERIALIZE_MINIMAL_THRESHOLD`, then the MazeDataset will use `serialize_minimal`.
# Setting to None means that `serialize_minimal` will never be used.
SERIALIZE_MINIMAL_THRESHOLD: int | None = 100


def set_serialize_minimal_threshold(threshold: int | None) -> None:
    global SERIALIZE_MINIMAL_THRESHOLD
    SERIALIZE_MINIMAL_THRESHOLD = threshold


def _load_maze_ctor(maze_ctor_serialized: str | dict) -> Callable:
    if isinstance(maze_ctor_serialized, dict):
        # this is both the new and old version of the serialization
        return GENERATORS_MAP[maze_ctor_serialized["__name__"]]
    elif isinstance(maze_ctor_serialized, str):
        # this is a version I switched to for a while but now we are switching back
        warnings.warn(
            f"you are loading an old model/config in `_load_maze_ctor()`!!! this should not be happening, please report: "
            + "https://github.com/understanding-search/maze-dataset/issues/new"
        )
        return GENERATORS_MAP[maze_ctor_serialized]
    else:
        raise ValueError(
            f"maze_ctor_serialized is of type {type(maze_ctor_serialized)}, expected str or dict"
        )


EndpointKwargsType = dict[
    typing.Literal[
        "except_when_invalid",
        "allowed_start",
        "allowed_end",
        "deadend_start",
        "deadend_end",
    ],
    bool | None | list[tuple[int, int]],
]


@serializable_dataclass(kw_only=True, properties_to_serialize=["grid_shape"])
class MazeDatasetConfig(GPTDatasetConfig):
    """maze dataset configuration, including tokenizers"""

    grid_n: int

    # not comparing n_mazes is done primarily to avoid conflicts which happen during `from_config` when we have applied filters
    n_mazes: int = serializable_field(compare=False)

    maze_ctor: Callable = serializable_field(
        default=GENERATORS_MAP["gen_dfs"],
        serialization_fn=lambda gen_func: {
            "__name__": gen_func.__name__,
            "__module__": gen_func.__module__,
            "__doc__": string_as_lines(gen_func.__doc__),
            "source_code": safe_getsource(gen_func),
        },
        loading_fn=lambda data: _load_maze_ctor(data["maze_ctor"]),
        assert_type=False,  # TODO: check the type here once muutils supports checking Callable signatures
    )

    maze_ctor_kwargs: dict = serializable_field(
        default_factory=dict,
        serialization_fn=lambda kwargs: kwargs,
        loading_fn=lambda data: (
            dict()
            if data.get("maze_ctor_kwargs", None)
            is None  # this should handle the backwards compatibility
            else data["maze_ctor_kwargs"]
        ),
    )

    endpoint_kwargs: EndpointKwargsType = serializable_field(
        default_factory=dict,
        serialization_fn=lambda kwargs: kwargs,
        loading_fn=lambda data: (
            dict()
            if data.get("endpoint_kwargs", None)
            is None  # this should handle the backwards compatibility
            else {
                k: (
                    # bools and Nones are fine
                    v
                    if (isinstance(v, bool) or v is None)
                    # assume its a CoordList
                    else [tuple(x) for x in v]  # muutils/zanj saves tuples as lists
                )
                for k, v in data["endpoint_kwargs"].items()
            }
        ),
        assert_type=False,
    )

    @property
    def grid_shape(self) -> CoordTup:
        return (self.grid_n, self.grid_n)

    @property
    def grid_shape_np(self) -> Coord:
        return np.array(self.grid_shape)

    @property
    def max_grid_n(self) -> int:
        return max(self.grid_shape)

    def stable_hash_cfg(self) -> int:
        return stable_hash(json.dumps(self.serialize()))

    def to_fname(self) -> str:
        return sanitize_fname(
            f"{self.name}-g{self.grid_n}-n{shorten_numerical_to_str(self.n_mazes)}-a_{self.maze_ctor.__name__.removeprefix('gen_')}-h{self.stable_hash_cfg()%10**5}"
        )

    def summary(self) -> dict:
        """return a summary of the config"""
        super_summary: dict = super().summary()
        self_ser: dict = self.serialize()
        return {
            **dict(
                name=self.name,
                fname=self.to_fname(),
                sdc_hash=self.stable_hash_cfg(),
                seed=self.seed,
                seq_len_min=self.seq_len_min,
                seq_len_max=self.seq_len_max,
                applied_filters=self.applied_filters,
            ),
            **{
                "grid_n": self_ser["grid_n"],
                "grid_shape": self_ser["grid_shape"],
                "n_mazes": self_ser["n_mazes"],
                "maze_ctor_name": self_ser["maze_ctor"]["__name__"],
                "maze_ctor_kwargs": self_ser["maze_ctor_kwargs"],
            },
        }


# TODO: don't use this unless generating in parallel!
def _generate_maze_helper(index: int) -> SolvedMaze:
    maze: LatticeMaze = _GLOBAL_WORKER_CONFIG.maze_ctor(
        grid_shape=_GLOBAL_WORKER_CONFIG.grid_shape_np,
        **_GLOBAL_WORKER_CONFIG.maze_ctor_kwargs,
    )
    solution = maze.generate_random_path(**_GLOBAL_WORKER_CONFIG.endpoint_kwargs)
    assert solution is not None, f"{solution = }"
    assert len(solution) > 0, f"{solution = }"
    assert isinstance(solution, np.ndarray), f"{solution = }"
    assert len(solution.shape) == 2, f"{solution = }"
    return SolvedMaze.from_lattice_maze(
        lattice_maze=maze,
        solution=solution,
    )


def _maze_gen_init_worker(config: MazeDatasetConfig):
    global _GLOBAL_WORKER_CONFIG
    _GLOBAL_WORKER_CONFIG = config

    # TODO
    # HACK: this makes the generation depend both on whether parallelism is used, and on the number of processes. this is bad!
    # only set numpy seed, since we do not use other random gens
    process_id: tuple[int] = multiprocessing.current_process()._identity
    if len(process_id) == 0:
        # no multiprocessing, seed was already set
        pass
    elif len(process_id) == 1:
        # multiprocessing, adjust seed based on process id
        np.random.seed(_GLOBAL_WORKER_CONFIG.seed + process_id[0])
    else:
        raise ValueError(
            f"unexpected process id: {process_id}\n{multiprocessing.Process()}"
        )


class MazeDataset(GPTDataset):
    """maze dataset"""

    def __init__(
        self,
        cfg: MazeDatasetConfig,
        mazes: typing.Sequence[SolvedMaze],
        generation_metadata_collected: dict | None = None,
    ) -> None:
        super().__init__()
        self.cfg: MazeDatasetConfig = cfg
        self.mazes: list[SolvedMaze] = list(mazes)
        self.generation_metadata_collected: dict | None = generation_metadata_collected

    def data_hash(self) -> int:
        return stable_hash(str(tuple([x.serialize() for x in self.mazes])))

    def __getitem__(self, i: int) -> SolvedMaze:
        return self.mazes[i]

    def __deepcopy__(self, memo) -> "MazeDataset":
        return MazeDataset.load(self.serialize())

    def as_tokens(
        self,
        maze_tokenizer,  # TODO: MazeTokenizer
        limit: int | None = None,
        join_tokens_individual_maze: bool = False,
    ) -> list[list[str]] | list[str]:
        """return the dataset as tokens

        if join_tokens_individual_maze is True, then the tokens of each maze are
        joined with a space, and the result is a list of strings.
        i.e.:
        >>> dataset.as_tokens(join_tokens_individual_maze=False)
        [["a", "b", "c"], ["d", "e", "f"]]
        >>> dataset.as_tokens(join_tokens_individual_maze=True)
        ["a b c", "d e f"]
        """
        output: list[list[str]] = [
            maze.as_tokens(maze_tokenizer) for maze in self.mazes[:limit]
        ]
        if join_tokens_individual_maze:
            return [" ".join(tokens) for tokens in output]
        else:
            return output

    def __len__(self) -> int:
        return len(self.mazes)

    def __eq__(self, other: typing.Any) -> bool:
        if not isinstance(other, MazeDataset):
            return NotImplemented
        # TODO: compare hashes of data instead of the data itself?
        return self.cfg == other.cfg and self.mazes == other.mazes

    @classmethod
    def generate(
        cls,
        cfg: MazeDatasetConfig,
        gen_parallel: bool = False,
        pool_kwargs: dict | None = None,
        verbose: bool = False,
    ) -> "MazeDataset":
        """generate a maze dataset"""

        # avoid copying since we dont want to pickle the staticmethod, just load/serialize it to avoid modifying the original config
        cfg_cpy = MazeDatasetConfig.load(cfg.serialize())

        if pool_kwargs is None:
            pool_kwargs = dict()
        mazes: list[SolvedMaze] = list()
        maze_indexes: Int[np.int8, "maze_index"] = np.arange(cfg_cpy.n_mazes)

        solved_mazes: list[SolvedMaze]
        tqdm_kwargs: dict = dict(
            total=cfg_cpy.n_mazes,
            unit="maze",
            desc="generating & solving mazes",
            disable=not verbose,
        )
        # TODO: don't use the global unless generating in parallel!
        if gen_parallel:
            with multiprocessing.Pool(
                **pool_kwargs,
                initializer=_maze_gen_init_worker,
                initargs=(cfg_cpy,),
            ) as pool:
                solved_mazes = list(
                    tqdm.tqdm(
                        pool.imap(
                            _generate_maze_helper,
                            maze_indexes,
                        ),
                        **tqdm_kwargs,
                    )
                )
        else:
            _maze_gen_init_worker(cfg_cpy)
            solved_mazes = list(
                tqdm.tqdm(
                    map(
                        _generate_maze_helper,
                        maze_indexes,
                    ),
                    **tqdm_kwargs,
                )
            )
        # reset seed to default value
        np.random.seed(cfg_cpy.seed)

        return cls(
            cfg=cfg_cpy,
            mazes=solved_mazes,
        )

    @classmethod
    def download(cls, cfg: MazeDatasetConfig, **kwargs) -> "MazeDataset":
        raise NotImplementedError("not implemented yet")

    @classmethod
    def load(cls, data: JSONitem) -> "MazeDataset":
        """load from zanj/json"""
        if data["__format__"] == "MazeDataset:minimal":
            return cls._load_minimal(data)
        elif data["__format__"] == "MazeDataset:minimal_soln_cat":
            return cls._load_minimal_soln_cat(data)
        else:
            assert data["__format__"] == "MazeDataset"
            return cls._load_full(data)

    @classmethod
    def _load_full(cls, data: JSONitem) -> "MazeDataset":
        assert data["__format__"] == "MazeDataset"
        return cls(
            cfg=MazeDatasetConfig.load(data["cfg"]),
            mazes=load_item_recursive(data["mazes"], tuple()),
            generation_metadata_collected=data["generation_metadata_collected"],
        )

    @classmethod
    def _load_minimal(cls, data: JSONitem) -> "MazeDataset":
        assert data["__format__"] == "MazeDataset:minimal"
        return cls(
            cfg=MazeDatasetConfig.load(data["cfg"]),
            generation_metadata_collected=data["generation_metadata_collected"],
            mazes=[
                SolvedMaze(
                    clist,
                    soln[:slen, ...],
                )
                for clist, slen, soln in zip(
                    load_item_recursive(data["maze_connection_lists"], tuple()),
                    load_item_recursive(data["maze_solution_lengths"], tuple()),
                    load_item_recursive(data["maze_solutions"], tuple()),
                    # load_item_recursive(data["maze_endpoints"], tuple()),
                )
            ],
        )

    @classmethod
    def _load_minimal_soln_cat(cls, data: JSONitem) -> "MazeDataset":
        assert data["__format__"] == "MazeDataset:minimal_soln_cat"

        maze_solution_lengths = load_item_recursive(
            data["maze_solution_lengths"], tuple()
        )
        maze_solutions_concat = load_item_recursive(
            data["maze_solutions_concat"], tuple()
        )
        maze_solutions = np.split(
            maze_solutions_concat, np.cumsum(maze_solution_lengths)[:-1], axis=0
        )

        return cls(
            cfg=load_item_recursive(data["cfg"], tuple()),
            generation_metadata_collected=load_item_recursive(
                data["generation_metadata_collected"], tuple()
            ),
            mazes=[
                SolvedMaze(
                    connection_list=clist,
                    solution=soln,
                )
                for clist, soln in zip(
                    load_item_recursive(data["maze_connection_lists"], tuple()),
                    # load_item_recursive(data["maze_endpoints"], tuple()),
                    maze_solutions,
                )
            ],
        )

    def serialize(self) -> JSONitem:
        """serialize to zanj/json"""
        if (
            SERIALIZE_MINIMAL_THRESHOLD is not None
            and len(self) >= SERIALIZE_MINIMAL_THRESHOLD
        ):
            return self._serialize_minimal()
        return self._serialize_full()

    def _serialize_full(self) -> JSONitem:
        return {
            "__format__": "MazeDataset",
            "cfg": json_serialize(self.cfg),
            "mazes": json_serialize(self.mazes),
            "generation_metadata_collected": json_serialize(
                self.generation_metadata_collected
            ),
        }

    def _serialize_minimal(self) -> JSONitem:
        if self.generation_metadata_collected is None:
            filtered_meta = self.filter_by.collect_generation_meta()
        else:
            filtered_meta = self

        max_solution_len: int = max(m.solution.shape[0] for m in filtered_meta.mazes)
        n_mazes: int = len(filtered_meta.mazes)
        grid_n: int = filtered_meta.cfg.grid_n

        maze_connection_lists: np.ndarray = np.empty(
            (n_mazes, 2, grid_n, grid_n), dtype=np.bool_
        )
        # maze_endpoints: np.ndarray = np.empty((n_mazes, 2, 2), dtype=np.int8)
        maze_solution_lengths: np.ndarray = np.empty((n_mazes,), dtype=np.int32)
        maze_solutions: np.ndarray = np.empty(
            (n_mazes, max_solution_len, 2), dtype=np.int8
        )

        for idx, maze in enumerate(filtered_meta.mazes):
            maze_connection_lists[idx] = maze.connection_list
            # maze_endpoints[idx] = np.array([maze.start_pos, maze.end_pos])
            maze_solution_lengths[idx] = maze.solution.shape[0]
            maze_solutions[idx, : maze.solution.shape[0]] = maze.solution

        return dict(
            __format__="MazeDataset:minimal",
            cfg=json_serialize(filtered_meta.cfg),
            generation_metadata_collected=json_serialize(
                filtered_meta.generation_metadata_collected
            ),
            maze_connection_lists=maze_connection_lists,
            # maze_endpoints=maze_endpoints,
            maze_solution_lengths=maze_solution_lengths,
            maze_solutions=maze_solutions,
        )

    def _serialize_minimal_soln_cat(self) -> JSONitem:
        if self.generation_metadata_collected is None:
            filtered_meta = self.filter_by.collect_generation_meta()
        else:
            filtered_meta = self

        maze_solution_lengths: np.ndarray = np.array(
            [m.solution.shape[0] for m in filtered_meta.mazes],
            dtype=np.int32,
        )
        n_mazes: int = len(filtered_meta.mazes)
        grid_n: int = filtered_meta.cfg.grid_n
        total_solution_len: int = np.sum(maze_solution_lengths)

        maze_connection_lists: np.ndarray = np.empty(
            (n_mazes, 2, grid_n, grid_n), dtype=np.bool_
        )
        maze_endpoints: np.ndarray = np.empty((n_mazes, 2, 2), dtype=np.int8)
        maze_solutions_concat: np.ndarray = np.empty(
            (total_solution_len, 2), dtype=np.int8
        )

        solutions_running_idx: int = 0
        for idx, maze in enumerate(filtered_meta.mazes):
            maze_connection_lists[idx] = maze.connection_list
            maze_endpoints[idx] = np.array([maze.start_pos, maze.end_pos])
            soln_len: int = maze.solution.shape[0]
            maze_solution_lengths[idx] = soln_len
            maze_solutions_concat[
                solutions_running_idx : solutions_running_idx + soln_len
            ] = maze.solution
            solutions_running_idx += soln_len

        return dict(
            __format__="MazeDataset:minimal_soln_cat",
            cfg=json_serialize(filtered_meta.cfg),
            generation_metadata_collected=json_serialize(
                filtered_meta.generation_metadata_collected
            ),
            maze_connection_lists=maze_connection_lists,
            maze_endpoints=maze_endpoints,
            maze_solution_lengths=maze_solution_lengths,
            maze_solutions_concat=maze_solutions_concat,
        )

    def update_self_config(self):
        """update the config to match the current state of the dataset"""
        self.cfg.n_mazes = len(self.mazes)

    def custom_maze_filter(
        self,
        method: typing.Callable[[SolvedMaze], bool],
        **kwargs,
    ) -> "MazeDataset":
        """filter the dataset using a custom method"""
        output: MazeDataset = MazeDataset(
            cfg=copy.deepcopy(self.cfg),
            mazes=[m for m in self.mazes if method(m, **kwargs)],
        )
        output.cfg.applied_filters.append(
            {
                "name": f"__custom__:{method.__name__}",
                "kwargs": kwargs,
            }
        )
        output.update_self_config()
        return output


MazeDatasetConfig._dataset_class = property(lambda self: MazeDataset)
register_loader_handler(
    LoaderHandler(
        check=lambda json_item, path=None, z=None: (
            isinstance(json_item, typing.Mapping)
            and "__format__" in json_item
            and json_item["__format__"].startswith("MazeDataset")
        ),
        load=lambda json_item, path=None, z=None: MazeDataset.load(json_item),
        uid="MazeDataset",
        source_pckg="maze_dataset.generation.maze_dataset",
        desc="MazeDataset",
    )
)


def register_maze_filter(
    method: typing.Callable[[SolvedMaze, typing.Any], bool]
) -> DatasetFilterProtocol:
    """register a maze filter, casting it to operate over the whole list of mazes

    method should be a staticmethod of a namespace class registered with `register_filter_namespace_for_dataset`

    this is a more restricted version of `register_dataset_filter` that removes the need for boilerplate for operating over the arrays
    """

    @functools.wraps(method)
    def wrapper(dataset: MazeDataset, *args, **kwargs):
        # copy and filter
        new_dataset: MazeDataset = copy.deepcopy(
            MazeDataset(
                cfg=dataset.cfg,
                mazes=[m for m in dataset.mazes if method(m, *args, **kwargs)],
            )
        )
        # update the config
        new_dataset.cfg.applied_filters.append(
            dict(name=method.__name__, args=args, kwargs=kwargs)
        )
        new_dataset.update_self_config()
        return new_dataset

    return wrapper


@register_filter_namespace_for_dataset(MazeDataset)
class MazeDatasetFilters:
    @register_maze_filter
    @staticmethod
    def path_length(maze: SolvedMaze, min_length: int) -> bool:
        """filter out mazes with a solution length less than `min_length`"""
        return len(maze.solution) >= min_length

    @register_maze_filter
    @staticmethod
    def start_end_distance(maze: SolvedMaze, min_distance: int) -> bool:
        """filter out datasets where the start and end pos are less than `min_distance` apart on the manhattan distance (ignoring walls)"""
        return np.linalg.norm(maze.start_pos - maze.end_pos, 1) >= min_distance

    @register_dataset_filter
    @staticmethod
    def cut_percentile_shortest(
        dataset: MazeDataset,
        percentile: float = 10.0,
    ) -> MazeDataset:
        """cut the shortest `percentile` of mazes from the dataset

        `percentile` is 1-100, not 0-1, as this is what `np.percentile` expects
        """
        lengths: np.ndarray = np.array([len(m.solution) for m in dataset])
        cutoff: int = int(np.percentile(lengths, percentile))

        filtered_mazes: list[SolvedMaze] = [
            m for m in dataset if len(m.solution) > cutoff
        ]
        new_dataset: MazeDataset = MazeDataset(cfg=dataset.cfg, mazes=filtered_mazes)

        return copy.deepcopy(new_dataset)

    @register_dataset_filter
    @staticmethod
    def truncate_count(
        dataset: MazeDataset,
        max_count: int,
    ) -> MazeDataset:
        """truncate the dataset to be at most `max_count` mazes"""
        new_dataset: MazeDataset = MazeDataset(
            cfg=dataset.cfg, mazes=dataset.mazes[:max_count]
        )
        return copy.deepcopy(new_dataset)

    @register_dataset_filter
    @staticmethod
    def remove_duplicates(
        dataset: MazeDataset,
        minimum_difference_connection_list: int | None = 1,
        minimum_difference_solution: int | None = 1,
        _max_dataset_len_threshold: int = 1000,
    ) -> MazeDataset:
        """remove duplicates from a dataset, keeping the **LAST** unique maze

        set minimum either minimum difference to `None` to disable checking

        if you want to avoid mazes which have more overlap, set the minimum difference to be greater

        Gotchas:
        - if two mazes are of different sizes, they will never be considered duplicates
        - if two solutions are of different lengths, they will never be considered duplicates
            TODO: check for overlap?
        """
        if len(dataset) > _max_dataset_len_threshold:
            raise ValueError(
                "this method is currently very slow for large datasets, consider using `remove_duplicates_fast` instead\n",
                "if you know what you're doing, change `_max_dataset_len_threshold`",
            )

        unique_mazes: list[SolvedMaze] = list()

        maze_a: SolvedMaze
        maze_b: SolvedMaze
        for i, maze_a in enumerate(dataset.mazes):
            a_unique: bool = True
            for maze_b in dataset.mazes[i + 1 :]:
                # after all that nesting, more nesting to perform checks
                if (minimum_difference_connection_list is not None) and (
                    maze_a.connection_list.shape == maze_b.connection_list.shape
                ):
                    if (
                        np.sum(maze_a.connection_list != maze_b.connection_list)
                        <= minimum_difference_connection_list
                    ):
                        a_unique = False
                        break

                if (minimum_difference_solution is not None) and (
                    maze_a.solution.shape == maze_b.solution.shape
                ):
                    if (
                        np.sum(maze_a.solution != maze_b.solution)
                        <= minimum_difference_solution
                    ):
                        a_unique = False
                        break

            if a_unique:
                unique_mazes.append(maze_a)

        return copy.deepcopy(
            MazeDataset(
                cfg=dataset.cfg,
                mazes=unique_mazes,
                generation_metadata_collected=dataset.generation_metadata_collected,
            )
        )

    @register_dataset_filter
    @staticmethod
    def remove_duplicates_fast(dataset: MazeDataset) -> MazeDataset:
        """remove duplicates from a dataset"""

        unique_mazes = list(dict.fromkeys(dataset.mazes))
        return copy.deepcopy(
            MazeDataset(
                cfg=dataset.cfg,
                mazes=unique_mazes,
                generation_metadata_collected=dataset.generation_metadata_collected,
            )
        )

    @register_dataset_filter
    @staticmethod
    def strip_generation_meta(dataset: MazeDataset) -> MazeDataset:
        """strip the generation meta from the dataset"""
        new_dataset: MazeDataset = copy.deepcopy(dataset)
        for maze in new_dataset:
            # hacky because it's a frozen dataclass
            maze.__dict__["generation_meta"] = None
        return new_dataset

    @register_dataset_filter
    @staticmethod
    def collect_generation_meta(
        dataset: MazeDataset,
        clear_in_mazes: bool = True,
        inplace: bool = True,
        allow_fail: bool = False,
    ) -> MazeDataset:
        if dataset.generation_metadata_collected is not None:
            return dataset
        # if the generation meta is already collected, don't collect it again, do nothing

        new_dataset: MazeDataset
        if inplace:
            new_dataset = dataset
        else:
            new_dataset = copy.deepcopy(dataset)

        gen_meta_lists: dict[bool | int | float | str | CoordTup, Counter] = (
            defaultdict(Counter)
        )
        for maze in new_dataset:
            if maze.generation_meta is None:
                if allow_fail:
                    break
                else:
                    raise ValueError(
                        "generation meta is not present in a maze, cannot collect generation meta"
                    )
            for key, value in maze.generation_meta.items():
                if isinstance(value, (bool, int, float, str)):
                    gen_meta_lists[key][value] += 1

                elif isinstance(value, set):
                    # special case for visited_cells
                    gen_meta_lists[key].update(value)

                elif isinstance(value, (list, np.ndarray)):
                    if isinstance(value, list):
                        try:
                            value = np.array(value)
                        except ValueError:
                            raise ValueError(
                                f"Cannot collect generation meta for {key} as it is a list of type '{str(type(value[0])) = }'",
                                "expected either a basic type (bool, int, float, str), a numpy coord, or a numpy array of coords",
                            )

                    if (len(value.shape) == 1) and (value.shape[0] == maze.lattice_dim):
                        # assume its a single coordinate
                        gen_meta_lists[key][tuple(value)] += 1
                    elif (len(value.shape) == 2) and (
                        value.shape[1] == maze.lattice_dim
                    ):
                        # assume its a list of coordinates
                        gen_meta_lists[key].update([tuple(v) for v in value])
                    else:
                        raise ValueError(
                            f"Cannot collect generation meta for {key} as it is an ndarray of shape {value.shape}",
                            "expected either a coord of shape (2,) or a list of coords of shape (n, 2)",
                        )
                else:
                    raise ValueError(
                        f"Cannot collect generation meta for {key} as it is of type '{str(type(value))}'",
                        "expected either a basic type (bool, int, float, str), a numpy coord, or a numpy array of coords",
                    )

            # clear the data
            if clear_in_mazes:
                # hacky because it's a frozen dataclass
                maze.__dict__["generation_meta"] = None

        new_dataset.generation_metadata_collected = {
            key: dict(value) for key, value in gen_meta_lists.items()
        }

        return new_dataset

        # the code below is for doing some smarter collecting and type checking. Probably will delete.
        """
        collect either the type at the field, or the shape of the field if it is an array
        metadata_types: dict[str, set[type, tuple]] = dict()
        for maze in new_dataset:
            for key, value in maze.generation_meta.items():
                if key not in metadata_types:
                    metadata_types[key] = set()

                if isinstance(value, np.ndarray):
                    metadata_types[key].add(value.shape)
                else:
                    metadata_types[key].add(type(value))

        # figure out what to do for each field
        metadata_actions: dict[str, typing.Callable] = dict()
        for key, key_type in metadata_types.items():
            if all(isinstance(kt, tuple) for kt in key_type):
                if all(kt == (2,) for kt in key_type):
                    # its all coords, do a statcounter on those coords
                    metadata_actions[key] = lambda vals: Counter(tuple(x) for x in vals)
                elif all(
                    (len(kt) == 2) and (kt[1] == 2) 
                    for kt in key_type
                ):
                    # its a list of coords, do a statcounter on those coords
                    metadata_actions[key] = lambda vals: Counter(
                        tuple(x) for x in np.concatenate(vals)
                    )
                else:
                    # its a list of something else, do a counter on those
                    # TODO: throw except here?
                    metadata_actions[key] = Counter
                    
            elif all(kt in (bool, int, float) for kt in key_type):
                # statcounter for numeric types
                metadata_actions[key] = StatCounter
            elif all(kt == str for kt in key_type):
                # counter for string types
                metadata_actions[key] = Counter
            else:
                # counter for everything else
                # TODO: throw except here?
                metadata_actions[key] = Counter
        """
